# Copyright 2021-2023 Nokia

class _Singleton(type):
    _instances = {}
    def __new__(cls, *args, **kwargs):
        res = super(_Singleton, cls).__new__(cls, *args, **kwargs)
        res.__copy__     = lambda self: self
        res.__deepcopy__ = lambda self, memo: self
        res.__reduce__   = lambda self: (self.__class__, ())
        return res

    def __call__(cls, *args, **kwargs):
        assert args == () and kwargs == {}
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class _Empty(metaclass=_Singleton):
    """Representation of whether empty leaf is present.

    .. Reviewed by TechComms 20210712
    """

    def __str__(self):
        return "Empty"

    def __repr__(self):
        return "Empty"

Empty = _Empty()
Empty.__doc__="""Define the YANG Empty type."""


# Copyright 2021-2023 Nokia

__all__ = ("SrosMgmtError", "InvalidPathError", "ModelProcessingError", "InternalError", "SrosConfigConflictError", "ActionTerminatedIncompleteError", "JsonDecodeError", "XmlDecodeError", )

__doc__ = """This module contains exceptions for error handling within pySROS.

.. reviewed by PLM 20211201
.. reviewed by TechComms 20211202
"""


class SrosMgmtError(Exception):
    """Exception raised by the :mod:`pysros.management` objects when:

    * data is missing
    * incorrect combinations of datastore and operation are performed
    * validation fails when using :py:meth:`pysros.management.Datastore.set`
    * invalid objects are passed to the operation e.g. :py:meth:`pysros.management.Datastore.set`

    .. Reviewed by PLM 20210625
    .. Reviewed by TechComms 20210630
    """
    pass

class InvalidPathError(Exception):
    """Exception raised when a path provided by the user:

    * is empty
    * fails to parse
    * does not point to an existing object
    * is missing list keys that must be provided

    .. Reviewed by PLM 20210625
    .. Reviewed by TechComms 20210630
    """
    pass

class ModelProcessingError(Exception):
    """Exception raised when an error occurs during processing of the YANG model (schema) when:

    * a YANG file cannot be found
    * a YANG file is malformed

    .. Reviewed by PLM 20210625
    .. Reviewed by TechComms 20210630
    """
    pass

class InternalError(Exception):
    """Exception raised for broader issues when:

    * schema creation fails and this unfinished schema is utilized

    .. Reviewed by PLM 20210625
    .. Reviewed by TechComms 20210630
    """
    pass

class SrosConfigConflictError(Exception):
    """Exception raised when a configuration commit failed due to conflicts
    between the ``candidate`` datastore and the ``baseline`` datastore.
    Retrying the configuration operation usually resolves the situation.
    If retrying does not resolve the issue, the connection should be closed using
    :py:meth:`pysros.management.Connection.disconnect` and the operation
    restarted to make a new connection.

    .. Reviewed by PLM 20211201
    .. Reviewed by TechComms 20211202
    """
    pass

class ActionTerminatedIncompleteError(Exception):
    """Exception raised when an operation (also known as an action)
    completes with a ``terminated-incomplete`` status.

    .. Reviewed by PLM 20211201
    .. Reviewed by TechComms 20211202
    """
    pass

class JsonDecodeError(Exception):
    """Exception raised when parsing json input fail."""
    pass

class XmlDecodeError(Exception):
    """Exception raised when parsing xml input fail."""
    pass

def make_exception(arg, **kwarg):
    """Create an exception.

    .. Reviewed by PLM 20211201
    .. Reviewed by TechComms 20211202
    """
    return arg[0](arg[1].format(**kwarg))



# Copyright 2021-2023 Nokia

import re

from .errors import *
from .singleton import _Singleton

class LazyBindModule(metaclass=_Singleton):
    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

class NoModule(metaclass=_Singleton):
    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

class Identifier:
    """Class to hold prefix-name pair for single YANG entry"""
    __slots__ = "_name", "_prefix"

    def __init__(self, module, name: str):
        assert ':' not in name
        if isinstance(module, str):
            assert ':' not in module
        self._name = name
        self._prefix = module

    @staticmethod
    def lazy_bound(name: str):
        return Identifier(LazyBindModule(), name)

    @staticmethod
    def builtin(name: str):
        return Identifier(NoModule(), name)

    def is_builtin(self):
        return self._prefix is NoModule()


    def is_lazy_bound(self):
        return self._prefix is LazyBindModule()

    @staticmethod
    def from_yang_string(s:str, default_module, prefix_module_mapping:dict):
        if ":" not in s:
            return Identifier(default_module, s)
        if s.count(":") != 1:
            raise make_exception(pysros_err_can_have_one_semicolon)
        prefix, name = s.split(":")
        if prefix not in prefix_module_mapping:
            raise make_exception(pysros_err_unknown_prefix_for_name, prefix=prefix, name=name)
        return Identifier(prefix_module_mapping[prefix], name)

    @staticmethod
    def from_model_string(s:str):
        if ":" not in s:
            return Identifier.lazy_bound(s)
        if s.count(":") != 1:
            raise make_exception(pysros_err_can_have_one_semicolon)
        return Identifier(*s.split(":"))

    def __hash__(self):
        return hash((self.prefix, self.name))

    def __str__(self):
        return  f"""{(str(self.prefix)+":") if self.prefix != NoModule() else ""}{self.name}"""

    def __repr__(self):
        if self.prefix is NoModule():
            return f"""Identifier.builtin({self.name!r})"""
        if self.prefix is LazyBindModule():
            return f"""Identifier.lazy_bound({self.name!r})"""
        return f"""Identifier({self.prefix!r}, {self.name!r})"""

    def __eq__(self, other):
        if type(other) is Identifier:
            return self._name == other._name and self._prefix == other._prefix
        elif type(other) == str:
            if ":" in other:
                return self == Identifier.from_model_string(other)
            return self.name == other
        return False

    def __ne__(self, other):
        return not(self == other)

    _IDENTIFIER_RE = re.compile(r'[a-zA-Z_][a-zA-Z0-9_\-.]*')
    def is_valid(self) -> bool:
        def valid_identifier(x):
            return bool(Identifier._IDENTIFIER_RE.fullmatch(x))
        return (valid_identifier(self._name) and
                (self._prefix is LazyBindModule()
                or self._prefix is NoModule()
                or valid_identifier(self._prefix)))

    @property
    def prefix(self):
        return self._prefix

    @property
    def name(self):
        return self._name

    @property
    def model_string(self):
        return f"{self.prefix+':' if self.prefix else ''}{self.name}"




# Copyright 2021-2023 Nokia

import collections
import re

from .errors import *


NL_RE = "[\n]+"
WSP_RE = "[ \t\r]+"
SINGLELINE_COMMENT_RE = "[/][/][^\n]*"
MULTILINE_COMMENT_RE = "[/][*].*?[*][/]"
UNQUOTED_STR_RE = """[^ \t\r\n"';{}](?:[^ \t\r\n;{}/]|[/*](?=[^/])|[/](?=[^/*]))*"""
QUOTED_STR_RE = """\
(?:["](?:[^\\\\"]|[\\\\][\\\\nt"])*["])\
|(?:['](?:[^'])*['])\
"""
STMT_END_RE = ";"
BLOCK_BEGIN_RE = "[{]"
BLOCK_END_RE = "[}]"

TOKEN_REGEX = re.compile(f"""(?:\
(?P<WSP>{WSP_RE})\
|(?P<NEW_LINES>{NL_RE})\
|(?P<SINGLELINE_COMMENT>{SINGLELINE_COMMENT_RE})\
|(?P<MULTILINE_COMMENT>{MULTILINE_COMMENT_RE})\
|(?P<UNQUOTED_STR>{UNQUOTED_STR_RE})\
|(?P<QUOTED_STR>{QUOTED_STR_RE})\
|(?P<STMT_END>{STMT_END_RE})\
|(?P<BLOCK_BEGIN>{BLOCK_BEGIN_RE})\
|(?P<BLOCK_END>{BLOCK_END_RE})\
|(?P<REST>.)\
)""", re.DOTALL)



TOKEN_KIND_STR = 0
TOKEN_STMT_END = 1
TOKEN_BLOCK_BEGIN = 2
TOKEN_BLOCK_END = 3
TOKEN_COMMENT = 4

TOKEN_NAME_TO_ID = {
    "UNQUOTED_STR": TOKEN_KIND_STR,
    "QUOTED_STR": TOKEN_KIND_STR,
    "STMT_END": TOKEN_STMT_END,
    "BLOCK_BEGIN": TOKEN_BLOCK_BEGIN,
    "BLOCK_END": TOKEN_BLOCK_END,
    "SINGLELINE_COMMENT_RE": TOKEN_COMMENT,
    "MULTILINE_COMMENT_RE": TOKEN_COMMENT,
}

def tokenize(s:str):
    for i in re.finditer(TOKEN_REGEX, s):
        if i.lastgroup in ("WSP", "SINGLELINE_COMMENT"):
            continue
        if i.lastgroup in ("NEW_LINES", "MULTILINE_COMMENT"):
            continue
        assert i.lastgroup != "REST"
        val = i.group()
        if i.lastgroup == "QUOTED_STR":
            val = val[1:-1]
        yield (TOKEN_NAME_TO_ID[i.lastgroup], val)

def yang_parser(yang, handler):
    it = iter(tokenize(yang))
    stack = collections.deque()
    ahead_token = None

    def get_token(expected):
        nonlocal ahead_token
        if ahead_token:
            res, ahead_token = ahead_token, None
        else:
            res = next(it, None)
        if res is None:
            if None in expected:
                return None
            raise make_exception(pysros_err_unexpected_end_of_yang)
        if res[0] not in expected:
            raise make_exception(pysros_err_unexpected_token, token=res[1])
        if res[0] == TOKEN_KIND_STR:
            ahead_token = next(it, None)
            while ahead_token and ahead_token[1] == "+":
                ahead_token = next(it, None)
                if ahead_token is None:
                    raise make_exception(pysros_err_unexpected_end_of_yang)
                if ahead_token[0] != TOKEN_KIND_STR:
                    raise make_exception(pysros_err_wrong_rhs)
                res = (res[0], res[1] + ahead_token[1])
                ahead_token = next(it, None)

        return res

    while True:
        token = get_token((TOKEN_KIND_STR, TOKEN_BLOCK_END, None, ))
        if not token:
            assert not stack
            break
        if token[0] == TOKEN_BLOCK_END:
            kw = stack.pop()
            handler.leave(kw)
            continue
        kw = token[1]
        token = get_token((TOKEN_KIND_STR, TOKEN_BLOCK_BEGIN, TOKEN_STMT_END, ))
        if token[0] == TOKEN_KIND_STR:
            arg = token[1]
            token = get_token((TOKEN_BLOCK_BEGIN, TOKEN_STMT_END, ))
        else:
            arg = None

        handler.enter(kw, arg)
        kind = token[0]
        if kind == TOKEN_STMT_END:
            handler.leave(kw)
        else:
            stack.append(kw)



# Copyright 2021-2023 Nokia

import operator

from abc import ABC, abstractmethod

from .errors import *
from .singleton import _Empty
from .yang_type import YangUnion, DECIMAL_LEAF_TYPE

__all__ = ("Action", "Container", "Leaf", "LeafList")

__doc__ = """This module contains wrappers describing the YANG 
structure and metadata obtained from SR OS.

.. Reviewed by PLM 20211201
.. Reviewed by TechComms 20211202
"""

class Schema:
    """YANG schema metadata information associated with elements in the data structure.

    .. note:: :py:class:`pysros.wrappers.Schema` metadata is read-only.

    .. property:: module

       YANG module name from which this node originates.

       :rtype: str

    .. property:: namespace

       YANG module namespace from which this node originates.  This is in URN or URL format.

       :rtype: str

    .. property:: yang_type

       YANG data type.  This type is the derived base YANG type, for example, if a YANG node uses
       a ``typedef``, and that ``typedef`` is a ``uint8``, yang_type returns ``uint8``.

       :rtype: str, :py:class:`SchemaType`

    .. property:: units

       The units defined in the YANG module.

       :rtype: str

    .. property:: default

       The default value as defined in the YANG module.

       :rtype: str, int

    .. property:: mandatory

       Identifies whether the item is required (mandatory) in the YANG module.

       :rtype: bool


    .. Reviewed by PLM 20221005
    .. Reviewed by TechComms 20221005
    """
    __slots__ = ("_model")
    _attributes = (
        "module",
        "namespace",
        "yang_type",
        "units",
        "default",
        "mandatory",
    )

    def __init__(self, model):
        self._model = model

    def __dir__(self):
        retval = super(self.__class__, self).__dir__()
        for attr in self._attributes:
            if hasattr(self, attr):
                retval.append(attr)
        return retval

    def __getattr__(self, attr):
        if attr == "module":
            return self._model.prefix
        if attr == "namespace":
            return self._model.namespace
        if attr == "yang_type":
            if self._model.yang_type is not None:
                return SchemaType(self._model.yang_type)
        if attr == "units":
            if self._model.units is not None:
                return self._model.units
        if attr == "default":
            if self._model.default is not None:
                return self._model.default
        if attr == "mandatory":
            if self._model.data_def_stm in (
                self._model.StatementType.leaf_,
                self._model.StatementType.leaf_list_,
                self._model.StatementType.choice_,
                self._model.StatementType.anydata_,
                self._model.StatementType.anyxml_,
            ):
                return True if self._model.mandatory else False
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __eq__(self, other):
        attrs_eq = lambda self, other, attr: getattr(self, attr, None) == getattr(other, attr, None)
        return (
            self.__class__ is other.__class__
            and all(attrs_eq(self, other, attr) for attr in self._attributes)
        )

class SchemaType:
    """Type information for YANG node. Resolves to a YANG base type.

    .. property:: range

       The range defined in YANG.

       :rtype: str

    .. container::

       .. property:: union_members

          Base YANG types that form part of the YANG union.

          :rtype: tuple

    .. Reviewed by PLM 20221005
    .. Reviewed by TechComms 20221005
    """
    __slots__ = ("_yang_type")
    _attributes = (
        "range",
        "union_members",
    )

    def __init__(self, yt):
        self._yang_type = yt

    def __dir__(self):
        retval = super(self.__class__, self).__dir__()
        for attr in self._attributes:
            if hasattr(self, attr):
                retval.append(attr)
        return retval

    def __getattr__(self, attr):
        if attr == "range":
            if self._yang_type.json_name() in DECIMAL_LEAF_TYPE:
                return self._yang_type.yang_range
        if attr == "union_members":
            if isinstance(self._yang_type, YangUnion):
                return tuple(SchemaType(yt) for yt in self._yang_type)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __str__(self):
        return self._yang_type.json_name()

    def __repr__(self):
        return self._yang_type.json_name()

    def __eq__(self, other):
        attrs_eq = lambda self, other, attr: getattr(self, attr, None) == getattr(other, attr, None)
        return (
            self.__class__ is other.__class__
            and self._yang_type.json_name() == other._yang_type.json_name()
            and all(attrs_eq(self, other, attr) for attr in self._attributes)
        )

class Wrapper:
    """Common functionality to support wrappers that describe the YANG structure from
    the SR OS schema.

    .. warning::
       Instance of this class SHOULD NOT be created by user of pysros library.

    .. Reviewed by PLM 20220923
    .. Reviewed by TechComms 20210712
    """

    __slots__ = ('_data', '_model')

    def __init__(self, value):
        self._data  = value
        self._model = None

    def __getattr__(self, attr):
        if attr == "_data":
            raise AttributeError()
        return getattr(self._data, attr)

    def __setattr__(self, attr, value):
        if attr == '_data' and not hasattr(self, '_data'):
            self._check_data_type(value)
            object.__setattr__(self, attr, value)
            return
        elif attr == '_model' and getattr(self, '_model', None) is None:
            object.__setattr__(self, attr, value)
            return
        raise AttributeError(f"'{self.__class__.__name__}' object attribute '{attr}' is read-only")

    def __delattr__(self, attr):
        raise make_exception(pysros_err_attr_cannot_be_deleted, obj=self.__class__.__name__, attribute=attr)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self._data == other._data

    @property
    def data(self):
        return self._data

    __module__ = 'pysros.wrappers'

    def __dir__(self):
        result = ['__module__', 'schema', 'data', '__getattr__', '__slots__']
        old_dir = object.__dir__(self)
        for item in dir(self.data):
            if not item.startswith('_') or item in old_dir:
                result.append(item)
        return result

    @property
    def schema(self):
        return Schema(self._model) if self._model else None

    @schema.setter
    def schema(self, _):
        raise make_exception(pysros_err_attr_is_read_only, obj=self.__class__.__name__, attribute = 'schema')

    @staticmethod
    @abstractmethod
    def _check_data_type(value):
        pass

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._data)


    @classmethod
    def _with_model(cls, data, model):
        obj = cls(data)
        obj._model = model
        return obj

def forward_methods(C, *methods):
    def make_forward(method):
        op = getattr(operator, method, None)
        if op is not None:
            # solves most methods like __add__, __and__, , but not special cases
            def forwad(self, *args):
                return op(self.data, *args)
            return forwad
        if method.startswith('__r'):
            n_method = "__" + method[3:]
            if hasattr(operator, n_method): # reverse operator __radd__, __rsub__, __rxor, etc.
                def forward(self, arg):
                    return getattr(operator, n_method)(arg, self.data)
                return forward

        # some methods (eg. __len__) are not defined in operator
        def forward(self, *args):
            return getattr(self.data, method)(*args)
        return forward
    for method in methods:
        setattr(C, method, make_forward(method))

class Container(Wrapper):
    """YANG container data structure node wrapper.

    A YANG container in the pySROS data structure behaves in the same way as a Python dict,
    where the value of a field can be obtained using the field name.

    :ivar data: dictionary containing the fields data

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """
    __slots__ = ()
    def __init__(self, value = None):
        super().__init__(value if value is not None else {})

    @classmethod
    def _check_data_type(cls, value):
        if not isinstance(value, dict):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name=cls.__name__)

forward_methods(Container, '__contains__', '__delitem__', '__getitem__', '__iter__', '__len__', '__reversed__', '__setitem__')

class Action(Container):
    """YANG Action data structure node wrapper.

    A YANG action in the pySROS data structure behaves in the same way as a pySROS :py:class:`Container`, but
    it is returned by the :py:meth:`pysros.management.Connection.convert` method.

    :ivar data: dictionary containing the fields data
    """
    pass


class LeafList(Wrapper):
    """YANG leaf-list data structure node wrapper.

    A YANG leaf-list in the pySROS data structure behaves in the same way as a Python list, where the separate
    values can be obtained using an index.

    :ivar data: list containing the values

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """
    __slots__ = ()

    def __init__(self, value = None):
        super().__init__(value if value is not None else [])

    @classmethod
    def _check_data_type(cls, values):
        if not isinstance(values, list):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name=cls.__name__)
        for value in values:
            Leaf._check_data_type(value)

forward_methods(LeafList, '__add__', '__contains__', '__delitem__', '__getitem__', '__iadd__', '__imul__', '__iter__', '__len__', '__mul__', '__reversed__', '__rmul__', '__setitem__')

class Leaf(Wrapper):
    """YANG leaf data structure node wrapper.

    Depending on the base (root) YANG type of the node, this object wraps a str, int, or bool.

    :ivar data: python object corresponding to the value

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210713
    """
    __slots__ = ()

    @staticmethod
    def _check_data_type(data):
        if not isinstance(data, (str, int, bool, _Empty)):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name = 'Leaf')

    def __bool__(self):
        return bool(self.data)

forward_methods(Leaf,
    '__lt__',
    '__le__',
    '__ge__',
    '__gt__',
    '__abs__',
    '__add__',
    '__and__',
    '__ceil__',
    '__contains__',
    '__divmod__',
    '__float__',
    '__floor__',
    '__floordiv__',
    '__getitem__',
    '__getnewargs__',
    '__index__',
    '__int__',
    '__invert__',
    '__iter__',
    '__len__',
    '__lshift__',
    '__mod__',
    '__mul__',
    '__neg__',
    '__or__',
    '__pos__',
    '__pow__',
    '__radd__',
    '__rand__',
    '__rdivmod__',
    '__rfloordiv__',
    '__rlshift__',
    '__rmod__',
    '__rmul__',
    '__ror__',
    '__round__',
    '__rpow__',
    '__rrshift__',
    '__rshift__',
    '__rsub__',
    '__rtruediv__',
    '__rxor__',
    '__sub__',
    '__truediv__',
    '__trunc__',
    '__xor__',
)
del forward_methods


# Copyright 2021-2023 Nokia

import copy
import xml.parsers.expat

from collections import defaultdict
from typing import Dict, DefaultDict, List, Set, Union, Optional

from .errors import *
from .identifier import Identifier, NoModule, LazyBindModule
from .model import AModel, Model, BuildingModel, StorageConstructionModel
from .model_path import ModelPath
from .model_walker import ModelWalker, DataModelWalker
from .tokenizer import yang_parser
from .yang_type import IdentityRef, LeafRef, PrimitiveType, UnresolvedIdentifier, YangType, YangTypeBase, YangUnion, Enumeration, Bits, should_be_buildin, type_from_name, DECIMAL_LEAF_TYPE, INTEGRAL_TYPES_MIN, INTEGRAL_TYPES_MAX, IP_TYPES

class YangHandler:
    """Handler for yang processing."""
    TAGS_WITH_MODEL = ("container", "list", "leaf", "typedef", "module", "submodule", "uses", "leaf-list", "import", "identity", "notification", "rpc", "input", "output", "choice", "case", "deviate", "action")
    TAGS_FORCE_MODULE = ("grouping", "identity", "typedef", "uses", "augment", "deviation", "type", "base")
    TAGS_ARG_IS_IDENTIFIER = ("base", "type", )
    TAGS_ARG_IS_YANG_PATH_NOT_ABSOLUTE_SCHEMA_ID = ("path", )
    assert not ({"config", "default", "fraction-digits", "length", "mandatory", "max-elements", "min-elements", "must", "range", "status", "type", "unique", "units", } & set(TAGS_WITH_MODEL)), "Substmt of deviate can not have model"
    TAGS_SHOULD_BE_PROCESSED = (
        "action",
        "anydata",
        "anyxml",
        "argument",
        "augment",
        "base",
        "belongs-to",
        "bit",
        "case",
        "choice",
        "config",
        "contact",
        "container",
        "default",
        "description",
        "deviate",
        "deviation",
        "enum",
#        "error-app-tag",
#        "error-message",
#        "extension",
        "feature",
        "fraction-digits",
        "grouping",
        "identity",
        "if-feature",
        "import",
        "include",
        "input",
        "key",
        "leaf",
        "leaf-list",
        "length",
        "list",
        "mandatory",
        "max-elements",
        "min-elements",
        "modifier",
        "module",
        "must",
        "namespace",
        "notification",
        "ordered-by",
        "organization",
        "output",
        "path",
        "pattern",
        "position",
        "prefix",
        "presence",
        "range",
        "reference",
        "refine",
        "require-instance",
        "revision",
        "revision-date",
        "rpc",
        "status",
        "submodule",
        "type",
        "typedef",
        "unique",
        "units",
        "uses",
        "value",
#        "when",
        "yang-version",
        "yin-element",
    )

    def __init__(self, builder: "ModelBuilder", root: BuildingModel):
        self.builder = builder
        self.root = root
        self.path = [root]
        self.full_path = []
        self.grouping_depth = 0
        self.in_type = 0
        self.derived_identities: Dict[Identifier, Set(Identifier)] = {}
        self.include_stack = [{}]
        self.module = None
        self.prefix = None
        self.ignore_depth = 0

    def enter(self, name, arg):
        if self.ignore_depth or name not in self.TAGS_SHOULD_BE_PROCESSED:
            self.ignore_depth += 1
            return
        self.full_path.append(name)

        if name in ("input", "output"):
            arg = name
        if name == "case" and not arg:
            arg = "unnamed"
        if name == 'typedef':
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            assert not new_id.is_lazy_bound()
            new = BuildingModel(new_id, BuildingModel.StatementType.typedef_, None)
            self.path.append(new)
        elif name in self.TAGS_WITH_MODEL:
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            new = BuildingModel(new_id, BuildingModel.StatementType[name.replace("-", "_") + "_"], self.model)
            self.path.append(new)
        elif name in ("augment", "deviation"):
            new_id = self.get_identifier(name, name in self.TAGS_FORCE_MODULE)
            new = BuildingModel(new_id, BuildingModel.StatementType[name + "_"], None)
            new.target_path = self.yang_path_to_model_path(arg)
            (self.builder.augments if name == "augment" else self.builder.deviations).append(new)
            self.path.append(new)
        elif name == "grouping":
            new_id = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            new = BuildingModel(new_id, BuildingModel.StatementType[name.replace("-", "_") + "_"], None)
            self.path.append(new)
            assert new_id not in self.builder.groupings
            self.builder.groupings[new_id] = new
        else:
            if name in self.TAGS_ARG_IS_IDENTIFIER:
                if name == "type" and should_be_buildin(arg):
                    arg = Identifier.builtin(arg)
                else:
                    arg = self.get_identifier(arg, name in self.TAGS_FORCE_MODULE)
            elif name in self.TAGS_ARG_IS_YANG_PATH_NOT_ABSOLUTE_SCHEMA_ID:
                arg = self.yang_path_to_model_path(arg, absolute_schema_id = False)
            self.model.blueprint.append((True, (name, arg)))

        handle_name = f'handle_early_{name.replace("-", "_DASH_")}'
        if hasattr(self, handle_name):
            handler = getattr(self, handle_name)
            handler(arg)

    def leave(self, name):
        if self.ignore_depth:
            self.ignore_depth -= 1
            return
        if name == "grouping":
            self.grouping_depth -= 1
        elif name == "typedef":
            self.construct(self.model)
            self.builder.types_to_resolve[self.model.name] = TypeDefModel(self.model)
        elif name in ("action", "rpc"):
            for child in self.model.children:
                if child.data_def_stm == BuildingModel.StatementType.input_:
                    break
            else:
                self.enter("input", None)
                self.leave("input")
            for child in self.model.children:
                if child.data_def_stm == BuildingModel.StatementType.output_:
                    break
            else:
                self.enter("output", None)
                self.leave("output")

        popped = self.full_path.pop()
        assert popped == name
        if name in (*self.TAGS_WITH_MODEL, "grouping", "augment", "deviation"):
            self.path.pop()
        else:
            self.model.blueprint.append((False, name))

    def construct(self, model=None):
        model = model or self.root
        self.path.append(model)
        self.full_path.append(model.data_def_stm.name.replace("_", "-")[:-1])
        for instruction in model.blueprint:
            if instruction[0]:
                self.full_path.append(instruction[1][0])
                if self.full_path[-1] == "type":
                    self.in_type += 1
                self.construct_enter(instruction[1])
            else:
                if self.full_path.pop() == "type":
                    self.in_type -= 1
        for child in self.model.children:
            self.construct(child)
        self.path.pop()
        self.full_path.pop()

    def construct_enter(self, instruction):
        name, arg = instruction
        handle_name = f'handle_{name.replace("-", "_DASH_")}'
        if hasattr(self, handle_name):
            handler = getattr(self, handle_name)
            handler(arg)

    def get_identifier(self, s, force_module=False):
        if self.module is not None:
            module = LazyBindModule() if (self.grouping_depth and not force_module) else self.module
            return Identifier.from_yang_string(s, module, self.prefixModuleMapping)
        return Identifier.builtin(s)

    @staticmethod
    def model_path_from_string(path: str, default_module: str, prefixModuleMapping: Dict[str, str], *,  absolute_schema_id):
        assert isinstance(path, str)
        if not path.startswith(('/', '../')):
            raise make_exception(pysros_err_cannot_pars_path, path=path)

        tokens = ModelWalker.tokenize_path(path, absolute_schema_id)
        result = ModelPath(Identifier.from_yang_string(
            p, default_module, prefixModuleMapping) for p in tokens)
        if not result.is_valid(only_absolute_path=absolute_schema_id):
            raise make_exception(pysros_err_invalid_yang_path, path=path)
        return result

    def yang_path_to_model_path(self, path, absolute_schema_id=True):
        default_module = self.module if absolute_schema_id else LazyBindModule()
        return self.model_path_from_string(path, default_module, self.prefixModuleMapping, absolute_schema_id=absolute_schema_id)

    @property
    def model(self):
        return self.path[-1]

    @property
    def model_type(self):
        return self.model.data_def_stm

    @property
    def prefixModuleMapping(self):
        return self.include_stack[-1]

    @property
    def last_yang_type(self):
        return self.model.yang_type[-1] if isinstance(self.model.yang_type, YangUnion) else self.model.yang_type

    def handle_identity(self, identifier):
        assert self.model.data_def_stm == BuildingModel.StatementType.identity_

    def handle_base(self, ident):
        if self.in_type:
            assert isinstance(self.last_yang_type, IdentityRef)
            self.last_yang_type.add_base(ident)
        elif self.model.data_def_stm == BuildingModel.StatementType.identity_:
            self.model.identity_bases.append(ident)

    def handle_path(self, arg: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, LeafRef)
            self.last_yang_type.set_path(arg)

    def handle_type(self, identifier: str):
        if not isinstance(self.model.yang_type, YangUnion):
            assert self.model.yang_type is None
            self.model.yang_type = type_from_name(identifier)
        elif (identifier != Identifier.builtin('union')):
            self.model.yang_type.append(type_from_name(identifier))

    def handle_units(self, arg: str):
        assert self.model.units is None
        self.model.units = arg

    def handle_namespace(self, arg: str):
        assert self.model.namespace is None
        self.model.namespace = arg

    def handle_default(self, arg: str):
        assert self.model.default is None
        self.model.default = arg

    def handle_mandatory(self, arg: str):
        assert self.model.mandatory is None
        self.model.mandatory = arg

    def handle_range(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.yang_range is None
        self.last_yang_type.yang_range = arg

    def handle_fraction_DASH_digits(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.fraction_digits is None
        self.last_yang_type.fraction_digits = int(arg)

    def handle_length(self, arg: str):
        assert isinstance(self.last_yang_type, (PrimitiveType, UnresolvedIdentifier))
        assert self.last_yang_type.length is None
        self.last_yang_type.length = arg

    def handle_status(self, arg: str):
        self.model.status = arg

    def handle_enum(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, Enumeration), f"expected Enumeration, got {self.last_yang_type}"
            self.last_yang_type.add_enum(value)

    def handle_bit(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, Bits)
            self.last_yang_type.add(value)

    def handle_value(self, value: str):
        if self.in_type:
            assert isinstance(self.last_yang_type, Enumeration)
            self.last_yang_type.set_last_enum_value(int(value))

    def handle_presence(self, _arg):
        self.model.presence_container = True

    def handle_ordered_DASH_by(self, arg):
        if arg == "user":
            self.model.user_ordered = True

    def handle_early_import(self, arg):
        self.builder.register_yang(arg)

    def handle_early_include(self, arg):
        self.include_stack.append({})
        self.builder.perform_parse(arg, self)
        self.include_stack.pop()

    def handle_key(self, arg):
        self.model.local_keys.extend(arg.split())

    def handle_early_module(self, arg):
        self.module = arg

    def handle_early_grouping(self, _arg):
        self.grouping_depth += 1

    def handle_early_prefix(self, arg):
        if self.full_path[-2] == "module":
            self.prefix = arg
        if self.full_path[-2] in ("import", "module"):
            assert arg not in self.prefixModuleMapping
            self.prefixModuleMapping[arg] = self.model.name.name

    def handle_config(self, arg: str):
        if arg not in ('false', 'true'):
            raise make_exception(pysros_err_invalid_config)
        self.model.config = (arg == 'true')

def _dummy_getter(filename):
    assert False, "Missing getter"

class ModelBuilder:
    """API for walking model tree."""
    def __init__(self, yang_getter=_dummy_getter, ns_map={}):
        self.root = BuildingModel("root", BuildingModel.StatementType["container_"], None)
        self.types_to_resolve = dict()
        self.groupings = dict()
        self.resolved_types = dict()
        self.all_yangs = set()
        self.parsed_yangs = set()
        self.files_to_parse = []
        self.augments = []
        self.deviations = []
        self.registered_modules = {}
        self._ns_map = ns_map
        self.yang_getter = yang_getter

    def get_module_content(self, yang_name):
        if yang_name in self.registered_modules:
            return self.registered_modules[yang_name]
        return self.yang_getter(yang_name)

    def register_yangs(self, yang_names):
        for name in yang_names:
            self.register_yang(name)

    def register_yang(self, yang_name):
        assert isinstance(yang_name, str)
        if yang_name not in self.all_yangs:
            self.all_yangs.add(yang_name)
            self.files_to_parse.append(yang_name)

    def DEBUG_parse_File(self, module_name, f):
        if module_name not in self.all_yangs:
            self.all_yangs.add(module_name)
            self.parsed_yangs.add(module_name)
        yang_parser(f, YangHandler(self, self.root))

    def DEBUG_register_module(self, module_name, f):
        assert module_name not in self.registered_modules
        self.registered_modules[module_name] = f

    def process_all_yangs(self):
        while self.files_to_parse:
            self.perform_parse(self.files_to_parse.pop())
        self.resolve()

    def resolve(self):
        self.resolve_groupings()
        self.resolve_augments()
        self.resolve_deviations()
        self.build_blueprints()
        self.resolve_typedefs()
        self.resolve_type_ranges()
        self.resolve_identities()
        self.resolve_leafrefs()
        self.resolve_config()
        self.resolve_namespaces()
        self.delete_blueprints()
        self.convert_model()

    def perform_parse(self, yang_name, yang_handler=None):
        if yang_name in self.parsed_yangs:
            return
        self.parsed_yangs.add(yang_name)

        yang_parser(self.get_module_content(yang_name), yang_handler or YangHandler(self, self.root))


    def add_child_to_uses(self, m: BuildingModel):
        if m.data_def_stm == BuildingModel.StatementType.uses_:
            assert not m.has_children
            i = m
            while True:
                if i.data_def_stm == BuildingModel.StatementType.module_:
                    module = i.name.name
                    break
                elif i.data_def_stm == BuildingModel.StatementType.augment_:
                    module = i.name.prefix
                    break
                i = i.parent

            def resolve_unresolved(m:BuildingModel):
                if m.name.prefix is LazyBindModule():
                    m.prefix = module

            for i in self.groupings[m.name].children:
                i.deepcopy(m)
            m.recursive_walk(resolve_unresolved)
            return False

    def set_correct_types(self, m:BuildingModel):
        if isinstance(m.yang_type, (UnresolvedIdentifier, YangUnion)):
            tdm = resolve_typedefs_deep(TypeDefModel(m), self.resolved_types)
            m.yang_type         = tdm.yang_type
            m.default           = tdm.default
            m.units             = tdm.units

    def resolve_typedefs(self):
        for key, value in self.types_to_resolve.items():
            assert not key in self.resolved_types
            self.resolved_types[key] = resolve_typedefs_deep(value, self.types_to_resolve)
            if key in IP_TYPES:
                assert isinstance(self.resolved_types[key].yang_type, YangUnion)
                self.resolved_types[key].yang_type.deduplicate()
                assert len(self.resolved_types[key].yang_type) == 1 and self.resolved_types[key].yang_type._types[0].identifier.name == "string"
                self.resolved_types[key].yang_type = self.resolved_types[key].yang_type._types[0]
        self.root.recursive_walk(self.set_correct_types)

    def set_correct_range(self, yt:YangTypeBase):
        assert yt.json_name() != "decimal64" or yt.fraction_digits
        if isinstance(yt, YangUnion):
            for t in yt:
                self.set_correct_range(t)

        if not yt or not yt.json_name() in (*DECIMAL_LEAF_TYPE, "string", "binary"):    return
        if yt.json_name() in ("string", "binary") and not yt.length:                    return

        full_range = False if (yt.yang_range or yt.json_name() in ("string", "binary")) else True
        work_range = yt.length if yt.json_name() in ("string", "binary") else yt.yang_range

        if full_range: work_range = "min..max"
        if yt.json_name() == "decimal64":
            min_val = str(-2**63+1)
            max_val = str(2**63-1)
            min_val = min_val[:-yt.fraction_digits] + '.' + min_val[-yt.fraction_digits:]
            max_val = max_val[:-yt.fraction_digits] + '.' + max_val[-yt.fraction_digits:]
        elif yt.json_name() in ("string", "binary"):
            min_val = str(0)
            max_val = str(2**64-1)
        else:
            min_val = INTEGRAL_TYPES_MIN[yt.json_name()]
            max_val = INTEGRAL_TYPES_MAX[yt.json_name()]

        work_range = work_range.replace("min", min_val)
        work_range = work_range.replace("max", max_val)

        if full_range:
            yt.yang_range = work_range
            return

        # merge adjacent subranges, such as "0|1..100" to "0..100" or "0..10|11..20" to "0..20"
        def process_subrange(subrange:str):
            nonlocal yt
            map_f       = float if yt.json_name() == "decimal64" else int
            i_subrange  = list(map(map_f, subrange.split("..")))
            assert len(i_subrange) in (1, 2)
            return i_subrange if len(i_subrange) == 2 else 2 * i_subrange

        subranges       = work_range.split("|")
        res_subranges   = [process_subrange(subranges[0])]
        for subrange in subranges[1:]:
            i_subrange = process_subrange(subrange)
            if res_subranges[-1][1] == i_subrange[0] - 1:
                res_subranges[-1][1] = i_subrange[1]
            else:
                res_subranges.append(i_subrange)

        res_subranges   = [subr if subr[0] != subr[1] else [subr[0]] for subr in res_subranges]
        work_range      = "|".join(["..".join(map(str, subr)) for subr in res_subranges])
        if yt.json_name() in ("string", "binary"):
            yt.length       = work_range
        else:
            yt.yang_range   = work_range

    def resolve_type_ranges(self):
        def setter(m:BuildingModel):
            if isinstance(m.yang_type, YangTypeBase):
                self.set_correct_range(m.yang_type)
        self.root.recursive_walk(setter)

    def resolve_identities(self):
        # first step creates dict, where we find for each identity
        # its all identities, that are directly derived from it.
        directly_derived: DefaultDict[Identifier, Set[Identifier]] = defaultdict(set)
        def find_derived(m:BuildingModel):
            if m.data_def_stm != BuildingModel.StatementType.identity_:
                return
            for b in m.identity_bases:
                directly_derived[b].add(m.name)

        self.root.recursive_walk(find_derived)
        # second step adds all direct and indirect derived identities together
        def add_derived_recursive(result_set, id):
            got = directly_derived.get(id, set())
            toUpdate = got - result_set
            result_set.update(got)
            for c in toUpdate:
                add_derived_recursive(result_set, c)

        derived: Dict[Identifier, Set(Identifier)] = {}
        for id in directly_derived.keys():
            derived[id] = set()
            add_derived_recursive(derived[id], id)

        # last step iterates through all identityrefs and set possible values
        def identityref_set_value(m:BuildingModel):
            if m.yang_type is None:
                return
            if type(m.yang_type) is IdentityRef:
                m.yang_type.set_values(derived)
            elif type(m.yang_type) is YangUnion:
                for st in m.yang_type:
                    if type(st) is IdentityRef:
                        st.set_values(derived)

        self.root.recursive_walk(identityref_set_value)

    def resolve_leafrefs(self):
        def replace_leafrefs(m: BuildingModel):
            if not isinstance(m.yang_type, LeafRef):
                return

            w = DataModelWalker(m)
            assert m.data_def_stm in (BuildingModel.StatementType.leaf_, BuildingModel.StatementType.leaf_list_)

            while isinstance(w.current.yang_type, LeafRef):
                path = w.current.yang_type.path
                module_name = w.current.name.prefix
                if path._path and path._path[0].name != '..':
                    # absolute path - go to root
                    while not w.is_root:
                        w.go_to_parent()
                w.go_to(path, module_name)
            m.yang_type = w.current.yang_type
        self.root.recursive_walk(replace_leafrefs)

    def resolve_groupings(self):
        def inner(m:BuildingModel):
            if m.data_def_stm == BuildingModel.StatementType.uses_:
                if m.has_children:
                    return False
                for child in self.groupings[m.name].children:
                    child.deepcopy(m)
        for grouping in self.groupings.values():
            grouping.recursive_walk(inner)
        self.root.recursive_walk(self.add_child_to_uses)
        for augment in self.augments:
            augment.recursive_walk(self.add_child_to_uses)

    def process_augment(self, m:BuildingModel):
        if m.data_def_stm == BuildingModel.StatementType.augment_:
            w = ModelWalker.path_parse(self.root, m.target_path)
            for i in m.children:
                i.deepcopy(w.current)

    def resolve_augments(self):
        current = self.augments
        while current:
            remaining = []
            for augment in current:
                w = ModelWalker(self.root)
                try:
                    w.go_to(augment.target_path)
                    node = w.current
                except InvalidPathError as e:
                    remaining.append(augment)
                    continue
                for i in augment.children:
                    i.deepcopy(node)
            if len(remaining) >= len(current):
                raise make_exception(pysros_err_unresolved_augment)
            current = remaining

    def filter_blueprint(self, function, blueprint):
        filtering_depth = 0
        for instruction in blueprint:
            if filtering_depth:
                filtering_depth += 1 if instruction[0] else -1
                continue
            if instruction[0]:
                if function(instruction[1]):
                    yield instruction
                else:
                    filtering_depth = 1
            else:
                yield instruction

    def resolve_deviations(self):
        for deviation in self.deviations:
            w = ModelWalker(self.root)
            w.go_to(deviation.target_path)
            for deviate in deviation.children:
                if deviate.name.name == "add":
                    w.current.blueprint.extend(deviate.blueprint)
                if deviate.name.name == "delete":
                    for instruction in deviate.blueprint:
                        if instruction[0]:
                            w.current.blueprint = list(self.filter_blueprint(lambda b: b[0:1] != instruction[1][0:1], w.current.blueprint))
                if deviate.name.name == "replace":
                    depth = 0
                    for instruction in deviate.blueprint:
                        if not depth:
                            w.current.blueprint = list(self.filter_blueprint(lambda b: b[0] != instruction[1][0], w.current.blueprint))
                        depth += 1 if instruction[0] else -1
                    w.current.blueprint = list(w.current.blueprint) + deviate.blueprint
                if deviate.name.name == "not-supported":
                    w.current.delete_from_parent(quiet=False)

    def resolve_namespaces(self):
        def ns_set(m:BuildingModel):
            if m.namespace or not m.name.prefix in self._ns_map.keys():
                return
            m.namespace = self._ns_map[m.name.prefix]
        for root_child in self.root.children:
            root_child.recursive_walk(ns_set)

    def build_blueprints(self):
        handler = YangHandler(self, self.root)
        handler.construct()

    def delete_blueprints(self):
        def deleter(m: BuildingModel):
            del m.blueprint
        self.root.recursive_walk(deleter)

    def resolve_config(self):
        def false_setter(m: BuildingModel):
            m.config = False
        def resolver(m: BuildingModel):
            if not m.config:
                m.recursive_walk(false_setter)
                return False

        self.root.recursive_walk(resolver)

    def convert_model(self):
        new_root = StorageConstructionModel("root", BuildingModel.StatementType["container_"], None)
        w = DataModelWalker(self.root)

        stack = [new_root]

        def enter_fnc(w: DataModelWalker):
            nonlocal stack
            new = StorageConstructionModel(w.get_name(), w.get_dds(), stack[-1])
            if isinstance(w.current.yang_type, YangUnion):
                w.current.yang_type.deduplicate()
            new.yang_type = w.current.yang_type
            new.units = w.current.units
            new.namespace = w.current.namespace
            new.default = w.current.default
            new.mandatory = w.current.mandatory
            new.status = w.current.status
            new.presence_container = w.current.presence_container
            new.user_ordered = w.current.user_ordered
            new.local_keys = w.current.local_keys
            new.target_path = w.current.target_path
            new.identity_bases = w.current.identity_bases
            new.config = w.current.config
            stack.append(new)

        def leave_fnc(w: DataModelWalker):
            nonlocal stack
            stack.pop()

        w.recursive_walk(enter_fnc=enter_fnc, leave_fnc=leave_fnc)
        self.root = Model(new_root._storage, new_root._index, None)

class TypeDefModel:
    __slots__ = ("yang_type", "default", "units")

    def __init__(self, arg: Union[YangType, Model, BuildingModel], default: Optional[str] = None, units: Optional[str] = None) -> None:
        assert isinstance(arg, (YangTypeBase, AModel))
        if isinstance(arg, AModel):
            self.yang_type          = arg.yang_type
            self.default            = arg.default
            self.units              = arg.units
        else:
            self.yang_type          = arg
            self.default            = default
            self.units              = units

    def __eq__(self, other):
        if not isinstance(other, self.__class__) or self.__slots__ != other.__slots__:
            return False
        return all(getattr(self, name) == getattr(other, name) for name in self.__slots__)

def merge_typedef_ranges(parent_range:str, child_range:str):
    """
    Vertically merge typedef integer ranges across typedef parent and child
    - keep child's integer range values (non-'min' or 'max')
    - replace child's 'min'/'max' strings with integer min or max values if present in parent
    - return child's ranges as-is if parent has no range defined or child has no 'max' or 'min'
    examples:
        - parent typedef range "1..200", child type range "min..100", result "1..100"
        - parent typedef range "1..200", child type range "100..max", result "100..200"
        - parent typedef range is None, child type range "100..max", result "100..max"
        - parent typedef range is '1..200', child type range "50..150", result "50..150"
    """
    if not parent_range:
        return child_range

    min_val     = parent_range.split("|")[0].split("..")[0]
    child_range = child_range.replace("min", min_val)
    max_val     = parent_range.split("|")[-1].split("..")[-1]
    child_range = child_range.replace("max", max_val)

    return child_range

def resolve_typedefs_shallow(t: TypeDefModel, typedefs: Dict[Identifier, TypeDefModel]):
    assert isinstance(t.yang_type, YangTypeBase)
    yang_type, default, units = t.yang_type, t.default, t.units
    while type(yang_type) is UnresolvedIdentifier and yang_type.identifier in typedefs:
        r_model = typedefs[yang_type.identifier]
        u_range = yang_type.yang_range
        frac_d  = yang_type.fraction_digits
        length  = yang_type.length
        if any((u_range, frac_d, length)):
            yang_type = copy.copy(r_model.yang_type)
            if u_range:
                yang_type.yang_range = merge_typedef_ranges(yang_type.yang_range, u_range)
            if frac_d:
                yang_type.fraction_digits = frac_d
            if length:
                yang_type.length = length
        else:
            yang_type = r_model.yang_type
        default = default   or r_model.default
        units   = units     or r_model.units

    assert not isinstance(yang_type, UnresolvedIdentifier), f"Cannot resolve type {yang_type}"
    assert isinstance(yang_type, YangTypeBase)
    return TypeDefModel(
        copy.copy(yang_type),
        default,
        units,
    )

def resolve_typedefs_deep(m: TypeDefModel, typedefs: Dict[Identifier, TypeDefModel]):
    m = resolve_typedefs_shallow(m, typedefs)
    if not isinstance(m.yang_type, YangUnion):
        return m

    u_yang_type, u_default, u_units = YangUnion(), None, None
    for sub in m.yang_type:
        tdm = resolve_typedefs_deep(TypeDefModel(sub), typedefs)
        if type(tdm.yang_type) == YangUnion:
            for t in tdm.yang_type:
                u_yang_type.append(t)
        else:
            u_yang_type.append(tdm.yang_type)
        u_default   = tdm.default
        u_units     = tdm.units
    m.yang_type         = u_yang_type
    m.default           = m.default         or u_default
    m.units             = m.units           or u_units
    return m



# Copyright 2021-2023 Nokia

import copy
import pprint
import json

from abc import ABC, abstractmethod
from collections import OrderedDict
from contextlib import ExitStack
from enum import Enum, auto
from lxml import etree
from ncclient.xml_ import to_ele
from typing import Union

from .errors import *
from .identifier import NoModule, Identifier
from .model import Model
from .model_walker import FilteredDataModelWalker, ModelWalker
from .singleton import _Singleton, Empty
from .wrappers import Wrapper, Action, Container, Leaf, LeafList

_get_tag_name = lambda x: etree.QName(x).localname
_text_in_tag_tail = lambda x: x.tail and x.tail.strip()
_text_in_tag_text = lambda x: x.text and x.text.strip()
_create_root_ele = lambda : etree.Element("dummy-root", nsmap={"nokia-attr": "urn:nokia.com:sros:ns:yang:sr:attributes"})

MO_STATEMENT_TYPES = (Model.StatementType.container_, Model.StatementType.list_, Model.StatementType.action_)
FIELD_STATEMENT_TYPES = (Model.StatementType.leaf_, Model.StatementType.leaf_list_)

def _raise_invalid_text_exception(tag, check_parent_tag=True):
    tag_to_check = tag.getparent() if check_parent_tag else tag
    if tag_to_check.tag == "dummy-root":
        raise make_exception(pysros_err_invalid_xml_root)
    else:
        raise make_exception(pysros_err_invalid_xml_element, element=_get_tag_name(tag_to_check))

def _get_local_ns_map(model, ns_map):
    if model.name.prefix in ns_map:
        return {None:ns_map[model.name.prefix]}
    elif model.name.prefix is NoModule():
        return {}
    else:
        raise make_exception(pysros_err_prefix_does_not_have_ns, prefix=model.name.prefix, name=model.name.name)

def _leaf_to_xml(value, root, walker, ns_map, replace_field=False):
    if walker.get_dds() == Model.StatementType.leaf_:
        value = (value, )
    for i in value:
        txt = walker.get_type().to_string(i)
        subelement(root, walker.current.name.name, _get_local_ns_map(walker.current, ns_map), txt, {"operation":"replace"} if replace_field else {})
        replace_field=False

class RequestData:
    """Basic API for holding and handling data.

    .. Reviewed by TechComms 20210712
    """

    class _Action(Enum):
        basic   = auto()
        convert = auto()

    def __init__(self, root:Model, ns_map:dict, action:_Action=_Action.basic, walker=FilteredDataModelWalker):
        self._root = root
        self._data = _ListStorage(root, self)
        self._ns_map = ns_map
        self._action = action
        self._Walker = walker

    def process_path(self, path:Union[str, FilteredDataModelWalker], *, strict=False):
        """Create all entries in given path and return setter for last section of the path.

        .. Reviewed by TechComms 20210712
        """
        walker = path if isinstance(path, ModelWalker) else self._Walker.user_path_parse(self._root, path, accept_root=(self._action == self._Action.convert))
        current = _ASetter.create_setter(self._data, self)
        for elem, keys in zip(walker.path, walker.keys):
            if not isinstance(current, _MoDataSetter):
                raise make_exception(pysros_err_missing_keys, element=current._walker.get_name())
            if elem.data_def_stm not in FIELD_STATEMENT_TYPES:
                if strict and not current.child_mos.is_created(elem.name.name):
                    raise make_exception(pysros_err_no_data_found)
                current = current.child_mos.get_or_create(elem.name.name)
                if keys:
                    if strict and not current.entry_exists_nocheck(keys):
                        raise make_exception(pysros_err_no_data_found)
                    current = current.entry_nocheck(keys)
            elif current.keys.can_contains(elem.name.name):
                current = current.keys.get(elem.name.name)
            else:
                if not current.fields.contains(elem.name.name):
                    if strict:
                        raise make_exception(pysros_err_no_data_found)
                    else:
                        current.fields.set_getValue(elem.name.name)
                current = current.fields.get(elem.name.name)

        return current

    def set_as_xml(self, value):
        """Populate storage from xml value.

        .. Reviewed by TechComms 20210712
        """
        d = value.xpath("/ncbase:rpc-reply/ncbase:data", namespaces={"ncbase": "urn:ietf:params:xml:ns:netconf:base:1.0"})
        if len(d) != 1:
            raise make_exception(pysros_err_wrong_netconf_response)
        d = d[0]

        root = _ASetter.create_setter(self._data, self)
        root.set_as_xml(d)

    def set_action_as_xml(self, path, value):
        data = value.xpath("/ncbase:rpc-reply", namespaces={"ncbase": "urn:ietf:params:xml:ns:netconf:base:1.0"})
        if len(data) != 1:
            raise make_exception(pysros_err_wrong_netconf_response)
        data = data[0]

        self.process_path(path).set_as_xml(data)

    def to_xml(self):
        """Return storage as xml.

        .. Reviewed by TechComms 20210712
        """
        root = etree.Element("root")
        self._data._to_xml(self._ns_map, root)
        return list(root[0])

    def xml_tag_has_correct_ns(self, tag, walker):
        """Check if xml namespace is correct.
        """
        if self._action == RequestData._Action.convert and etree.QName(tag).namespace:
            prefix = walker.get_child(_get_tag_name(tag)).current.prefix
            if prefix not in self._ns_map or self._ns_map[prefix] != etree.QName(tag).namespace:
                return False
        return True

    def debug_to_model(self):
        """Print storage as model.

        .. Reviewed by TechComms 20210712
        """
        model = self._data.to_model()
        pprint.pprint(model)

    def debug_dump(self):
        """Dump rquest data in raw form.

        .. Reviewed by TechComms 20210712
        """
        self._data.debug_dump()

    def _unwrap(self, value):
        return value.data if isinstance(value, Wrapper) else value


class _AStorage(ABC):
    """Abstract representation of data storage.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, rd:RequestData):
        self.rd = rd

    def to_xml(self, ns_map, root):
        """Return data in xml format.

        .. Reviewed by TechComms 20210712
        """
        return self._to_xml(ns_map, root)

    @abstractmethod
    def _to_xml(self, ns_map, root):
        """Implementation of returning data in xml format.

        """
        pass

    @abstractmethod
    def to_model(self, *, key_filter={}):
        """Return data as user model.

        .. Reviewed by TechComms 20210712
        """
        pass

    @property
    def _walker(self):
        return self.rd._Walker(self._model)

    def _resolve_xml_name(self, model, ns_map):
        if model.name.prefix in ns_map:
            return etree.QName(ns_map[model.name.prefix], model.name.name)
        elif model.name.prefix is NoModule():
            return model.name.name
        else:
            raise make_exception(pysros_err_prefix_does_not_have_ns, prefix=model.name.prefix, name=model.name.name)

def subelement(parent, tag, local_nsmap, text=None, attrib=None):
    """Wrapper around etree.Subelement, that raises SrosMgmtError instead of ValueError."""
    try:
        result = etree.SubElement(parent, tag, attrib, nsmap=local_nsmap)
        if text is not None:
            result.text = text
        return result
    except ValueError as err:
        raise SrosMgmtError(err.args) from None

class _MoStorage(_AStorage):
    """Data storage for containers and list entries (list+local keys).

    .. Reviewed by TechComms 20210712
    """

    def __init__(self, model:Model, local_keys, rd:RequestData):
        super().__init__(rd)
        self._model = model
        self._local_keys = copy.deepcopy(local_keys)
        self._child = {}
        self._operation = ""
        self._replace_field = "" # Could be field storage

    def _to_xml(self, ns_map, root):
        root_attr = None if not self._operation else {"operation": self._operation}
        root = subelement(root, self._model.name.name, _get_local_ns_map(self._model, ns_map), None, root_attr)
        for k, v in self._local_keys.items():
            if v is FieldValuePlaceholder():
                pass
            elif v is GetValuePlaceholder():
                self._leaf_placeholder_to_xml(v, root, self._walker.get_child(k), ns_map)
            else:
               txt  = self._walker.get_child(k).get_type().to_string(v)
               subelement(root, self._walker.get_child(k).current.name.name, _get_local_ns_map(self._model, ns_map), txt)

        for k, v in self._child.items():
            if isinstance(v, (_MoStorage, _ListStorage)):
                v._to_xml(ns_map, root)
            else:
                if v is Delete():
                    subelement(root, self._walker.get_child(k).current.name.name, _get_local_ns_map(self._model, ns_map), None, {"operation": "remove"})
                elif v is FieldValuePlaceholder():
                    pass
                elif v is GetValuePlaceholder():
                    self._leaf_placeholder_to_xml(v, root, self._walker.get_child(k), ns_map)
                elif self._replace_field and self._replace_field == k:
                    _leaf_to_xml(v, root, self._walker.get_child(k), ns_map, replace_field=True)
                else:
                    _leaf_to_xml(v, root, self._walker.get_child(k), ns_map)

    def to_xml(self, ns_map, root):
        root_attr = None if not self._operation else {"operation": self._operation}

        for k, v in self._child.items():
            if isinstance(v, (_MoStorage, _ListStorage)):
                v._to_xml(ns_map, root)
            else:
                if v is Delete():
                    subelement(root, self._walker.get_child(k).current.name.name, _get_local_ns_map(self._model, ns_map), None, {"operation": "remove"})
                elif v is FieldValuePlaceholder():
                    pass
                elif v is GetValuePlaceholder():
                    self._leaf_placeholder_to_xml(v, root, self._walker.get_child(k), ns_map)
                elif self._replace_field and self._replace_field == k:
                    _leaf_to_xml(v, root, self._walker.get_child(k), ns_map, replace_field=True)
                else:
                    _leaf_to_xml(v, root, self._walker.get_child(k), ns_map)

    def _leaf_placeholder_to_xml(self, value, root, walker, ns_map):
        subelement(root, walker.current.name.name, _get_local_ns_map(walker.current, ns_map))

    def to_model(self, *, key_filter={}):
        data = {}
        is_selection_filter = {} in key_filter.values()
        for k, v in self._local_keys.items():
            if is_selection_filter and k not in key_filter:
                continue
            data[k] = self._leaf_to_model(k, v)
        for k, v in self._child.items():
            key_proxy = DictionaryKeysProxy(self.rd._unwrap(key_filter))
            if is_selection_filter and self._walker.get_child(k).current.name not in key_proxy:
                continue
            if not isinstance(v, (_MoStorage, _ListStorage)):
                data[k] = self._leaf_to_model(k, v)
            else:
                data[k] = v.to_model(key_filter=(key_filter.get(k, {})))
                if not data[k] and not self._walker.get_child(k).has_explicit_presence():
                    del data[k]
        if self._walker.is_root:
            return data
        if self._model.data_def_stm == Model.StatementType.action_:
            return Action._with_model(data, self._model)
        return Container._with_model(data, self._model)

    def _leaf_to_model(self, name, value):
        assert self.rd._action == RequestData._Action.convert or value is not GetValuePlaceholder()
        assert value is not FieldValuePlaceholder()
        model = self._walker.get_child(name).current
        if self._walker.get_child_dds(name) == Model.StatementType.leaf_:
            return Leaf._with_model(value, model)
        return LeafList._with_model(value, model)

    def keys_equal(self, keys):
        """Compare if keys are equal. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        return self._local_keys == keys

    def get_keys_flat(self):
        keys = tuple(self._local_keys[key] for key in self._model.local_keys)
        if len(keys) == 1: keys = keys[0]
        return keys

    def debug_dump(self, indent=0):
        line = " "*indent
        if self._operation:
            line += self._operation + " "
        line += self._walker.get_name().name
        if self._local_keys:
            line += " ["
            line += ", ".join(f"{k}={v}" for k, v in self._local_keys.items())
            line += "]"
        line += " {"
        print(line)
        for k, v in self._child.items():
            if isinstance(v, _AStorage):
                v.debug_dump(indent+1)
            else:
                print(f"{' '*(indent+1)}{k} = {v}")
        print((" "*indent)+"}")


class _ListStorage(_AStorage):
    """Storage for all entries for a specific list.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, model:Model, rd:RequestData):
        if model.data_def_stm in (Model.StatementType.container_, Model.StatementType.action_):
            self.__class__ = _MoStorage
            self.__class__.__init__(self, model, {}, rd)
        else:
            super().__init__(rd)
            self._model = model
            self._entries = {}
            self._operation = ""

    def _to_xml(self, ns_map, root):
        if self._entries:
            for entry in self._entries.values():
                entry._to_xml(ns_map, root)
        else:
            subelement(root, self._model.name.name, _get_local_ns_map(self._model, ns_map))

    def to_xml(self, ns_map, root):
        if self._entries:
            for entry in self._entries.values():
                entry._to_xml(ns_map, root)

    def to_model(self, *, key_filter={}):
        d = OrderedDict() if self._model.user_ordered else {}
        for e in self._entries.values():
            d[e.get_keys_flat()] = e.to_model(key_filter=key_filter)

        return d

    def get_entry(self, keys):
        """Get entry by keys. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        res = self._entries.get(tuple(sorted(keys.items())), None)
        if res is None:
            raise make_exception(pysros_err_entry_does_not_exists)
        return res

    def get_or_create_entry(self, keys):
        """Get entry by keys. Create the entry if it does not exist. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        try:
            return self.get_entry(keys)
        except KeyError:
            return self.add_entry(keys)

    def has_entry(self, keys):
        """Determine if entry with specific keys exists. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        try:
            self.get_entry(keys)
            return True
        except KeyError:
            return False

    def add_entry(self, keys):
        """Add new entry with specific keys. Entry cannot already exist. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        assert not self.has_entry(keys)
        res = _MoStorage(self._model, keys, self.rd)
        self._entries[tuple(sorted(keys.items()))] = res
        return res

    def debug_dump(self, indent=0):
        for i in self._entries.values():
            i.debug_dump(indent+1)

class DictionaryKeysProxy():
    """Proxy for dictionary keys which unwraps keys and compares by performing Identifier.__eq__"""
    def __init__(self, data):
        if not isinstance(data, dict):
            raise make_exception(pysros_err_invalid_value, data=data)
        self.data = data

    def __getitem__(self, key:Identifier):
        #need to call Identifier.__eq__
        for k in self.data.keys():
            if key == k:
                return self._unwrap(self.data[k])
        raise KeyError(key)

    def __contains__(self, val:Identifier):
        #need to call Identifier.__eq__
        for ele in self.data:
            if val == ele:
                return True
        return False

    def __str__(self):
        return str(self.data)

    def _unwrap(self, val):
        return val.data if isinstance(val, Wrapper) else val

class RdJsonEncoder(json.JSONEncoder):
    def add_ns(self, o, d, walker:ModelWalker):
        if o is self.root:
            return {walker.get_child(k).get_name().model_string: v for k, v in d.items()}
        return {walker.get_child(k).get_name().model_string if walker.get_name().prefix != walker.get_child(k).get_name().prefix else k: v for k, v in d.items()}

    def default(self, o):
        is_root = o is self.root
        if isinstance(o, _MoStorage):
            res = {}
            if not is_root:
                res.update(self.add_ns(o, o._local_keys, o._walker))
            elif is_root and self.force_key:
                res[o._walker.get_child(self.force_key).get_name().model_string] = o._local_keys[self.force_key]
            res.update(self.add_ns(o, o._child, o._walker))
            stringify = lambda x, dds: [str(v) for v in x] if dds == Model.StatementType.leaf_list_ else str(x)
            cvt = lambda k, v: stringify(v, o._walker.get_child_dds(k)) if o._walker.get_child_dds(k) in FIELD_STATEMENT_TYPES and o._walker.get_child_type(k).json_name() in ("int64", "uint64") else v
            res = {k: cvt(k, v) for k, v in res.items()}
            return res
        elif isinstance(o, _ListStorage):
            res = list(o._entries.values())
            return {o._walker.get_name().model_string: res} if is_root else res
        elif isinstance(o, (str, list, dict, bool)):
            return o
        elif o is Empty:
            return [None]
        else:
            return str(o)

class _ASetter(ABC):
    """Data setter representation for storage classes.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, storage:"_AStorage", rd:RequestData):
        self._storage = storage
        self.rd = rd

    @property
    def _walker(self):
        return self.rd._Walker(self._storage._model)

    @staticmethod
    def create_setter(storage:"_AStorage", rd:RequestData):
        if isinstance(storage, _MoStorage):
            return _MoDataSetter(storage, rd)
        else:
            return _ListSetter(storage, rd)

    @abstractmethod
    def set(self, value):
        """Set data in model format.

        .. Reviewed by TechComms 20210712
        """
        pass

    @abstractmethod
    def set_as_xml(self, value):
        """Set data in xml format.

        .. Reviewed by TechComms 20210712
        """
        pass

    def checkJsonDuplicates(x):
        if len(x) != len({i[0] for i in x}):
            raise make_exception(pysros_err_multiple_occurences_of_node)
        return dict(x)

    def set_as_json(self, value):
        """Set data in json format."""
        try:
            val = json.loads(value, object_pairs_hook=_ASetter.checkJsonDuplicates)
        except json.JSONDecodeError as e:
            raise JsonDecodeError(*e.args) from None

        if not isinstance(val, dict):
            raise make_exception(pysros_err_wrong_json_root)

        self._set_as_json(val, is_root = True)

    @abstractmethod
    def _set_as_json(self, value, is_root = False):
        """Implementation of setter for JSON."""
        pass

    def to_json(self, pprint = True):
        """Return data in json format."""
        members = {
            "root": self._storage,
            "root_setter": self,
            "force_key": getattr(self, "_key_name", None),
        }
        encoder = type("RdJsonEncoderSpecialization", (RdJsonEncoder, ), members)
        indent = 4 if pprint else None
        return json.dumps(self._storage, cls=encoder, indent=indent)

    @abstractmethod
    def delete(self):
        """Set data operation to delete"""
        pass

    @abstractmethod
    def replace(self):
        """Set data operation to replace"""
        pass

    def entry_get_keys(self):
        """Set list keys as a placeholders"""
        raise make_exception(pysros_err_target_should_be_list)


    def to_model(self, *, key_filter={}):
        """Return data in model format.

        .. Reviewed by TechComms 20210712
        """
        return self._storage.to_model(key_filter=key_filter)

    def to_xml(self):
        """Return data in xml format."""
        root = _create_root_ele()
        self._storage.to_xml(self.rd._ns_map, root)
        return root


    def set_filter(self, value):
        """Populate request data with filter syntax"""
        raise make_exception(pysros_err_filter_not_supported_on_leaves)

    def is_compare_supported_endpoint(self):
        return False

    def is_action(self):
        return self._walker.get_dds() == Model.StatementType.action_

    def entry_xml(self, value):
        """For XML there is no tag for the list itself, only for entries.
        Therefore, the list setter can be avoided, however, you must create
        an entry setter and it simplifies the code if you always create
        an entry regardless of the child type.  It is effective only for
        the list setter."""
        return self

    def _as_storage_type(self, val, *, child_name=None):
        walker = self._walker.get_child(child_name) if child_name else self._walker
        if walker.get_dds() == Model.StatementType.leaf_list_:
            return [walker.get_type().as_storage_type(v, self.rd._action == RequestData._Action.convert) for v in val]
        else:
            return walker.get_type().as_storage_type(val, self.rd._action == RequestData._Action.convert)

    def _handle_entry_keys_namespaces(self, entry):
        """Strip namespace prefixes from entry key names.
        Also raise an error if there are two identical entry keys, one with the namespace prefix and other one without it.
        """
        local_keys = [Identifier(self._walker.get_name().prefix, k) for k in self._walker.get_local_key_names()]
        for k in local_keys:
            if k.model_string in entry:
                if k.name in entry:
                    raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=entry[k.model_string])
                entry[k.name] = entry.pop(k.model_string)

class _LeafSetter(_ASetter):
    """Interface for managing leafs. Because leafs do not have dedicated storage, the
       class has its own implementation of to_model method.

    .. Reviewed by TechComms 20210713
    """
    def __init__(self, storage:"_ListStorage", leaf_name:str, rd:RequestData):
        super().__init__(storage, rd)
        self._leaf_name = leaf_name

    def set(self, value):
        value = self.rd._unwrap(value)
        if not self._walker.check_field_value(value):
            raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._leaf_name)
        else:
            self._set_nocheck(self._as_storage_type(value))

    def _set_nocheck(self, value):
        self._storage._child[self._leaf_name] = value

    def set_getValue(self):
        self._storage._child[self._leaf_name] = GetValuePlaceholder()

    def set_as_xml(self, value):
        if not len(value):
            raise make_exception(pysros_err_malformed_xml)
        for v in value:
            self.set_or_append_as_xml(v)

    def _set_as_json(self, value, is_root = False):
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_json_structure)
        value = {k:v for k, v in value.items() if not k.startswith("@")}
        if len(value) != 1:
            raise make_exception(pysros_err_invalid_json_structure)

        val = next(iter(value.items()))
        if self._walker.get_name() != val[0]:
            raise make_exception(pysros_err_invalid_json_structure)
        val = val[1]
        if self._walker.get_type().json_name() in ("int64", "uint64"):
            if self._walker.get_dds() != Model.StatementType.leaf_list_:
                val = [self.rd._unwrap(val)]
            else:
                val = self.rd._unwrap(val)
            if not isinstance(val, list) or not all(isinstance(v, str) for v in val):
                raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._walker.get_name().name)
            try:
                value = [int(self.rd._unwrap(v)) for v in val]
            except:
                raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._walker.get_name().name) from None
        self.set(self._walker.as_model_type(val))

    def set_or_append_as_xml(self, value_element):
        if _text_in_tag_tail(value_element) or (self._walker.get_dds() == Model.StatementType.leaf_list_ and _text_in_tag_text(value_element.getparent())):
            _raise_invalid_text_exception(value_element)
        value = self._walker.as_model_type(value_element.text or "")

        if self._walker.get_dds() == Model.StatementType.leaf_list_ and isinstance(self._storage._child.get(self._leaf_name), list):
            value = self._storage._child[self._leaf_name] + value
        self.set(value)

    def to_model(self, *, key_filter={}):
        return self._storage._leaf_to_model(self._leaf_name, self._storage._child[self._leaf_name])

    def to_xml(self):
        root = _create_root_ele()
        _leaf_to_xml(self._storage._child[self._leaf_name], root, self._walker, self.rd._ns_map)
        return root

    def delete(self):
        if self._walker.get_dds() == Model.StatementType.leaf_list_:
            raise make_exception(pysros_err_invalid_operation_on_leaflist)
        self._storage._child[self._leaf_name] = Delete()

    def replace(self):
        self._storage._replace_field = self._leaf_name

    @property
    def _walker(self):
        return super()._walker.get_child(self._leaf_name)

class _KeySetter(_ASetter):
    """Interface for keys. Most of this class are stub methods to provide
       setter interface to keys."""
    def __init__(self, storage:"_ListStorage", key_name:str, rd:RequestData):
        super().__init__(storage, rd)
        self._key_name = key_name

    def set(self, value):
        value = self.rd._unwrap(value)
        if not self._walker.check_field_value(value):
            raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._key_name)
        elif self._as_storage_type(value) != self._storage._local_keys[self._key_name]:
            raise make_exception(pysros_err_key_val_mismatch, key_name=self._key_name)

    def set_getValue(self):
        pass

    def set_placeholder(self):
        pass

    def to_model(self, *, key_filter={}):
        return self._storage._leaf_to_model(self._key_name, self._storage._local_keys[self._key_name])

    def to_xml(self):
        root = _create_root_ele()
        _leaf_to_xml(self._storage._local_keys[self._key_name], root, self._walker, self.rd._ns_map)
        return root

    def delete(self):
        raise make_exception(pysros_err_invalid_operation_on_key)

    def replace(self):
        raise make_exception(pysros_err_invalid_operation_on_key)

    def set_as_xml(self, value):
        if _text_in_tag_tail(value):
            _raise_invalid_text_exception(value_element)
        if not len(value):
            raise make_exception(pysros_err_malformed_xml)
        for v in value:
            if _text_in_tag_tail(v):
                _raise_invalid_text_exception(v)
            self.set(self._walker.as_model_type(v.text or ""))

    def _set_as_json(self, value, is_root = False):
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_json_structure)
        value = {k:v for k, v in value.items() if not k.startswith("@")}
        if  len(value) != 1:
            raise make_exception(pysros_err_invalid_json_structure)

        val = next(iter(value.items()))
        if self._walker.get_name() != val[0]:
            raise make_exception(pysros_err_invalid_json_structure)
        if self._walker.get_type().json_name() in ("int64", "uint64"):
            if not isinstance(val[1], str):
                raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._walker.get_name().name)
            try:
                value = int(val[1])
            except:
                raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=self._walker.get_name().name) from None
        self.set(self._walker.as_model_type(val[1]))

    @property
    def _walker(self):
        return super()._walker.get_child(self._key_name)

class _ListSetter(_ASetter):
    """Interface for managing instances of specific list.

    .. Reviewed by TechComms 20210712
    """

    def set(self, value):
        """Set a value in a list.

        Keys may be in format {(key1, key2) : Container({}), (key1, key2) : Container({})}
        or directly in entry {key1: value1, key2: value2, field1 : value}.

        .. Reviewed by PLM 20211018
        """
        is_wrapped = isinstance(value, Container)
        value = self.rd._unwrap(value)
        value = copy.copy(value)
        unwrapper = DictionaryKeysProxy(value)
        if value and not is_wrapped and self._walker.dict_keys(value):
            for k, v in value.items():
                self._handle_entry_keys_namespaces(v)
                self.entry(self._tuple_to_dict(k)).set(v)
        elif self._walker.entry_keys(unwrapper):
            self._handle_entry_keys_namespaces(value)
            self.entry(value).set(value)
        else:
            raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=value)

    def set_filter(self, value):
        value = copy.copy(value)
        self._handle_entry_keys_namespaces(value)
        for k in self._walker.get_local_key_names():
            if k not in value:
                value[k] = FieldValuePlaceholder()
        self.entry_nocheck(value).set_filter(value)

    def _tuple_to_dict(self, t):
        return {k:v for k, v in zip(self._walker.get_local_key_names(), (t if isinstance(t, tuple) else (t, )))}

    def _extract_keys(self, entry):
        return {k:self._as_storage_type(v, child_name=k) for k, v in entry.items() if k in self._walker.get_local_key_names()}

    def _convert_keys_to_model(self, entry):
        try:
            def unwrap(v):
                return v.data if isinstance(v, Wrapper) else v
            val =  {k: (GetValuePlaceholder() if unwrap(v) in (GetValuePlaceholder(), {}) else self._walker.as_child_model_type(k, unwrap(v))) for k, v in entry.items() if v is not FieldValuePlaceholder()}
            return val
        except:
            raise make_exception(pysros_err_invalid_key_in_path) from None

    def _check_and_unwrap_keys(self, entry, *, json=False):
        for k in self._walker.get_local_key_names():
            if k not in entry:
                raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=k)
            entry[k] = self.rd._unwrap(entry[k])
            if not self._walker.check_child_field_value(k, entry[k], json=json):
                raise make_exception(pysros_err_incorrect_leaf_value, leaf_name=k)
            entry[k] = self._walker.get_child_type(k).as_storage_type(entry[k], self.rd._action == RequestData._Action.convert)

    def entry(self, value, *, json=False):
        """Receive entry with specified keys in value. Keys are expected as dict(name->value). Additional
        fields may be present (no verification for extra fields).

        .. Reviewed by TechComms 20210712
        """
        if json:
            value = copy.copy(value)
        self._handle_entry_keys_namespaces(value)
        self._check_and_unwrap_keys(value, json=json)
        return _MoDataSetter(self._storage.get_or_create_entry(self._extract_keys(value)), self.rd)

    def add_entry(self, value, *, json=False):
        if json:
            value = copy.copy(value)
        self._handle_entry_keys_namespaces(value)
        self._check_and_unwrap_keys(value, json=json)
        keys = self._extract_keys(value)
        if self._storage.has_entry(keys):
            raise make_exception(pysros_err_multiple_occurences_of_entry)
        return _MoDataSetter(self._storage.get_or_create_entry(keys), self.rd)

    def entry_nocheck(self, value):
        """Receive entry with specified keys in value without checking for the correct type. Keys are expected
        as dict(name->value). Additional fields may be present (no verification for extra fields).

        .. Reviewed by PLM 20211018
        """
        return _MoDataSetter(self._storage.get_or_create_entry(self._convert_keys_to_model(self._extract_keys(value))), self.rd)

    def entry_exists_nocheck(self, value):
        """Receive entry with specified keys in value without checking for the correct type. Keys are expected
        as dict(name->value). Additional fields may be present (no verification for extra fields)
        Returns true if entry exists, otherwise false.
        """
        return self._storage.has_entry(self._convert_keys_to_model(self._extract_keys(value)))

    def entry_xml(self, value):
        keys = {}
        for e in value:
            if _get_tag_name(e) in self._walker.get_local_key_names():
                if _text_in_tag_tail(e):
                    _raise_invalid_text_exception(e)
                keys[_get_tag_name(e)] = e.text or ""
        if set(keys.keys()) != set(self._walker.get_local_key_names()):
            raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=keys)
        if self.entry_exists_nocheck(keys):
            raise make_exception(pysros_err_multiple_occurences_of_entry)
        return self.entry_nocheck(keys)

    def set_as_xml(self, value):
        for v in value:
            self.entry_xml(v).set_as_xml(v)

    def _set_as_json(self, value, is_root = False):
        if is_root:
            if not isinstance(value, dict):
                raise make_exception(pysros_err_invalid_json_structure)
            value = {k:v for k, v in value.items() if not k.startswith("@")}
            if len(value) != 1:
                raise make_exception(pysros_err_invalid_json_structure)
            value = next(iter(value.items()))
            if self._walker.get_name() != value[0] or not isinstance(value[1], list):
                raise make_exception(pysros_err_invalid_json_structure)
            value = value[1]
        else:
            if not isinstance(value, list) or not all(isinstance(i, dict) for i in value):
                raise make_exception(pysros_err_convert_invalid_value_for_type, name=self._walker.get_name().name)
        for v in value:
            self.add_entry(v, json=True)._set_as_json(v)

    def delete(self):
        raise make_exception(pysros_err_invalid_path_operation_missing_keys)

    def replace(self):
        for k, v in self._storage._entries.items():
            v._operation = "replace"

    def entry_get_keys(self):
        keys = {key:GetValuePlaceholder() for key in self._walker.get_local_key_names()}
        setter = _MoDataSetter(self._storage.get_or_create_entry(keys), self.rd)
        setter.set({})
        return setter

class _MoDataSetter(_ASetter):
    """Interface managing specific list entry or container.

    .. Reviewed by TechComms 20210712
    """
    class _AChild:
        def __init__(self, setter:"_MoDataSetter"):
            self._setter = setter

        @property
        def _walker(self):
            return self._setter.rd._Walker(self._setter._storage._model)

    class _Keys(_AChild):
        """Interface for retrieving and setting keys.

        .. Reviewed by TechComms 20210712
        """
        def set(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name).set(value)

        def set_as_json(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name)._set_as_json({name: value})

        def set_getValue(self, name):
            self.get(name).set_getValue()

        def set_placeholder(self, name):
            self.get(name).set_placeholder()

        def get(self, name):
            if not self.can_contains(name):
                raise make_exception(pysros_err_unknown_child, child_name=name, path=self._walker._get_path())
            return _KeySetter(self._setter._storage, name, self._setter.rd)

        def contains(self, name):
            return self.can_contains(name)

        def can_contains(self, name):
            return self._walker.has_local_key_named(name)

    class _Fields(_AChild):
        """Interface for retrieving and setting fields.

        .. Reviewed by TechComms 20210712
        """

        def set(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name).set(value)

        def set_as_json(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name)._set_as_json({name: value})

        def set_getValue(self, name):
            self.get(name).set_getValue()

        def get(self, name):
            if not self.can_contains(name):
                raise make_exception(pysros_err_unknown_child, child_name=name, path=self._walker._get_path())
            return _LeafSetter(self._setter._storage, name, self._setter.rd)

        def get_nonexisting(self, name):
            if not self.can_contains(name):
                raise make_exception(pysros_err_unknown_child, child_name=name, path=self._walker._get_path())
            if self.contains(name) and self._walker.get_child_dds(name) != Model.StatementType.leaf_list_:
                raise make_exception(pysros_err_multiple_occurences_of_node)
            return _LeafSetter(self._setter._storage, name, self._setter.rd)

        def contains(self, name):
            return name in self._setter._storage._child

        def can_contains(self, name):
            return self._walker.has_field_named(name)


    class _ChildMos(_AChild):
        """Interface for retrieving and setting children.

        .. Reviewed by TechComms 20210712
        """
        def set(self, name, value):
            self.get_or_create(Identifier.from_model_string(name).name).set(value)

        def set_as_json(self, name, value):
            self.get_or_create(Identifier.from_model_string(name).name)._set_as_json(value)

        def get_or_create(self, name):
            if self._walker.get_child_dds(name) in MO_STATEMENT_TYPES:
                if not self.is_created(name):
                    self._setter._storage._child[name] = _ListStorage(self._setter._walker.get_child(name).current, self._setter.rd)
                return _ASetter.create_setter(self._setter._storage._child[name], self._setter.rd)
            else:
                raise KeyError(name)

        def get(self, name):
            if self._walker.get_child_dds(name) in MO_STATEMENT_TYPES:
                if self.is_created(name):
                    if not self._walker.get_child_dds(name) == Model.StatementType.list_:
                        raise make_exception(pysros_err_multiple_occurences_of_node)
                else:
                    self._setter._storage._child[name] = _ListStorage(self._setter._walker.get_child(name).current, self._setter.rd)
                return _ASetter.create_setter(self._setter._storage._child[name], self._setter.rd)
            else:
                raise KeyError(name)

        def is_created(self, name):
            return self._walker.get_child_dds(name) in MO_STATEMENT_TYPES and name in self._setter._storage._child


    @property
    def keys(self):
        return self._Keys(self)

    @property
    def fields(self):
        return self._Fields(self)

    @property
    def child_mos(self):
        return self._ChildMos(self)

    def set(self, value):
        value = self.rd._unwrap(value)
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_value, data=value)
        children_to_set = set()
        walker = self._walker
        for k, v in value.items():
            if k in walker.get_local_key_names():
                self.keys.set(k, v)
            elif walker.get_child_dds(k) in FIELD_STATEMENT_TYPES:
                self.fields.set(k, v)
            elif walker.get_child_dds(k) in MO_STATEMENT_TYPES:
                self.child_mos.set(k, v)
            else:
                raise make_exception(pysros_err_unknown_dds, dds=walker.get_child_dds(k))
            name = Identifier.from_model_string(k).name
            if name in children_to_set:
                raise make_exception(pysros_err_duplicate_found, duplicate=name)
            children_to_set.add(name)

    def set_filter(self, value):
        value = self.rd._unwrap(value)
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_value, data=value)
        self._handle_entry_keys_namespaces(value)
        children_to_set = set()
        for k, v in value.items():
            v = self.rd._unwrap(v)
            if k in self._walker.get_local_key_names():
                if v == {} or v is GetValuePlaceholder():
                    self.keys.set_getValue(k)
                elif v is FieldValuePlaceholder():
                    self.keys.set_placeholder(k)
                else:
                    self.keys.set(k, v)
            elif self._walker.get_child_dds(k) in FIELD_STATEMENT_TYPES:
                if v == {}:
                    self.fields.set_getValue(k)
                else:
                    self.fields.set(k, v)
            elif self._walker.get_child_dds(k) in MO_STATEMENT_TYPES:
                self.child_mos.get_or_create(k).set_filter(v)
            else:
                raise make_exception(pysros_err_unknown_dds, dds=self._walker.get_child_dds(k))
            name = Identifier.from_model_string(k).name
            if name in children_to_set:
                raise make_exception(pysros_err_duplicate_found, duplicate=name)
            children_to_set.add(name)

    def set_as_xml(self, value):
        walker = self._walker
        if self.rd._action != RequestData._Action.convert:
            walker.return_blocked_regions = True
        if _text_in_tag_text(value):
            _raise_invalid_text_exception(value, check_parent_tag=False)
        if _text_in_tag_tail(value):
            _raise_invalid_text_exception(value)
        for e in value:
            if not walker.has_child(_get_tag_name(e)) or not self.rd.xml_tag_has_correct_ns(e, walker):
                raise make_exception(pysros_err_unknown_child, child_name=_get_tag_name(e), path=walker._get_path())
            if walker.is_region_blocked_in_child(_get_tag_name(e)):
                continue
            else:
                if walker.get_child_dds(_get_tag_name(e)) in FIELD_STATEMENT_TYPES:
                    if _get_tag_name(e) not in walker.get_local_key_names():
                        self.fields.get_nonexisting(_get_tag_name(e)).set_or_append_as_xml(e)
                    elif self.rd._action == RequestData._Action.convert:
                        xml = to_ele(f"<{_get_tag_name(value)}>{etree.tostring(e, encoding='unicode')}</{_get_tag_name(value)}>")
                        _KeySetter(self._storage, _get_tag_name(e), self.rd).set_as_xml(xml)
                elif walker.get_child_dds(_get_tag_name(e)) in MO_STATEMENT_TYPES:
                    self.child_mos.get(_get_tag_name(e)).entry_xml(e).set_as_xml(e)

    def _set_as_json(self, value, is_root = False):
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_json_structure)

        children_to_set = set()
        for k, v in value.items():
            if k.startswith("@"):
                continue
            self._walker.get_child(k) #module name check?
            k = Identifier.from_model_string(k).name
            if k in self._walker.get_local_key_names():
                self.keys.set_as_json(k, v)
            elif self._walker.get_child_dds(k) in FIELD_STATEMENT_TYPES:
                self.fields.set_as_json(k, v)
            elif self._walker.get_child_dds(k) in MO_STATEMENT_TYPES:
                self.child_mos.set_as_json(k, v)
            else:
                raise make_exception(pysros_err_unknown_dds, dds=self._walker.get_child_dds(k))
            name = Identifier.from_model_string(k).name
            if name in children_to_set:
                raise make_exception(pysros_err_duplicate_found, duplicate=name)
            children_to_set.add(name)

    def delete(self):
        self._storage._operation = "delete"

    def replace(self):
        self._storage._operation = "replace"

    def is_compare_supported_endpoint(self):
        return True

class GetValuePlaceholder(metaclass=_Singleton):
    pass

class FieldValuePlaceholder(metaclass=_Singleton):
    pass

class Delete(metaclass=_Singleton):
    pass



# Copyright 2021 Nokia

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='pysros',
    version='22.10.6',
    packages=['pysros'],
    url='https://www.nokia.com',
    license='Copyright 2021-2023 Nokia.  License available in the LICENSE.md file.',
    author='Nokia',
    author_email='',
    description='Python for the Nokia Service Router Operating Systems (pySROS)',
    project_urls={
        "Documentation": "https://network.developer.nokia.com/static/sr/learn/pysros/latest/",
        "Source": "https://github.com/nokia/pysros",
    },
    classifiers=[
        "License :: Other/Proprietary License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Internet",
        "Development Status :: 5 - Production/Stable",
    ],
    install_requires=[
        "ncclient~=0.6.12",
        "lxml~=4.9.2",
    ],
    python_requires=">=3.10",
    long_description=long_description,
    long_description_content_type="text/markdown",
)




# Copyright 2021-2023 Nokia

import copy
import locale
import sys

from enum import Enum, IntFlag, IntEnum, auto
from typing import Any, Dict, List, Optional, Tuple

from lxml import etree

from .errors import *
from .identifier import Identifier, NoModule
from .yang_type import Enumeration, YangType, YangTypeBase
from .yang_type import YangType

class AModel:
    class StatementType(Enum):
        leaf_         = 0
        list_         = 1
        container_    = 2
        leaf_list_    = 3
        choice_       = 4
        case_         = 5
        augment_      = 6
        uses_         = 7
        typedef_      = 8
        module_       = 9
        submodule_    = 10
        grouping_     = 11
        import_       = 12
        identity_     = 13
        action_       = 14
        anydata_      = 15
        anyxml_       = 16
        notification_ = 17
        rpc_          = 18
        input_        = 19
        output_       = 20
        deviation_    = 21
        deviate_      = 22

    __slots__ = ()

    def recursive_walk(self, cb):
        cont = cb(self)
        if cont == False:
            return
        for i in self.children:
            i.recursive_walk(cb)

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
       return f"""Model("{str(self.name)}")"""

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def debug_print(self, prefix="", last = False):
        class Colors:
            RED    = '\033[1;31;48m'
            GREEN  = '\033[1;32;48m'
            DGREEN = '\033[2;32;48m'
            BLUE   = '\033[1;34;48m'
            CYAN   = '\033[1;36;48m'
            PURPLE = '\033[1;35;48m'
            YELLOW = '\033[1;33;48m'
            RESET  = '\033[1;37;0m'

        utf_supported = locale.getlocale()[1] == "UTF-8"
        class AsciiArt:
            CHILD =         "+-- " if not utf_supported else "\u251c\u2500\u2500 "
            LAST_CHILD =    "+-- " if not utf_supported else "\u2514\u2500\u2500 "
            VERTICAL_LINE = "   |" if not utf_supported else "   \u2502"

        data_def_stm = self.data_def_stm.name[:-1]

        if sys.stdout.isatty():
            colorize = lambda data, color : color + str(data) + Colors.RESET
        else:
            colorize = lambda data, color : str(data)

        if not prefix:
            field_prefix = ""
        elif last:
            field_prefix = AsciiArt.LAST_CHILD
        else:
            field_prefix = AsciiArt.CHILD
        if data_def_stm == "container":
            t = colorize(data_def_stm, Colors.YELLOW)
        elif data_def_stm == "list":
            t = colorize(data_def_stm, Colors.GREEN)
        else:
            t = colorize(data_def_stm, Colors.CYAN)
        t_sufix = []
        if data_def_stm in ("leaf", "typedef", ):
            t_sufix.append(colorize(self.yang_type, Colors.BLUE))
        elif data_def_stm == "identity":
            bases = [str(b) for b in self.identity_bases]
            if bases:
                t_sufix.append(colorize(f"base=[{','.join(bases)}]", Colors.BLUE))
        elif data_def_stm == "list":
            t_sufix.append(colorize(",".join(self.local_keys), Colors.DGREEN))
            if self.user_ordered:
                t_sufix.append(colorize("user ordered", Colors.PURPLE))
        elif data_def_stm == "container":
            if self.presence_container:
                t_sufix.append(colorize("presence", Colors.PURPLE))
        elif data_def_stm == "typedef":
            t_sufix.append(colorize(f"->{self.yang_type}", Colors.YELLOW))
        elif data_def_stm == "augment":
            t_sufix.append(colorize(self.target_path, Colors.PURPLE))

        print(f"""{prefix[:-1]}{field_prefix}{self.debug_flags()} {self.name} [{" ".join((t,*t_sufix))}]""")

        remain = self.children_size - 1
        for i in self.children:
            new_prefix = prefix + ("    " if not remain else AsciiArt.VERTICAL_LINE)
            if not prefix:
                new_prefix = new_prefix[3:]
            i.debug_print(new_prefix, not remain)
            remain -= 1

    def debug_flags(self) -> str:
        """Return text flags:
            rw  for configuration data
            ro  for non-configuration data
            -x  for rpcs and actions
            -n  for notifications
            mp   for schema mount points

        .. Reviewed by TechComms 20210713
        """
        data_def_stm = self.data_def_stm.name[:-1]
        if data_def_stm in ("leaf", "list", "leaf_list", "container", "choice", "case"):
            return {
                True:  'rw',
                False: 'ro',
                None:  '??'
            }[self.config]
        if data_def_stm in ("rpc", "action"):
            return "-x"
        if data_def_stm == "notification":
            return "-n"
        return "mp"

    def test_print(self, prefix="", _last = False):
        data_def_stm = self.data_def_stm.name[:-1]
        if not prefix:
            field_prefix = ""
        else:
            field_prefix = "+-- "
        t_sufix = [data_def_stm]
        if data_def_stm in ("leaf", "typedef"):
            t_sufix.append(str(self.yang_type))
        elif data_def_stm == "list":
            t_sufix.append(",".join(self.local_keys))
            if self.user_ordered:
                t_sufix.append("user ordered")
        elif data_def_stm == "container":
            if self.presence_container:
                t_sufix.append("presence")
        elif data_def_stm == "typedef":
            t_sufix.append(f"->{self.yang_type}")
        elif data_def_stm == "augment":
            t_sufix.append(self.target_path)

        res = f"""{prefix[:-1]}{field_prefix}{self.name} [{" ".join(t_sufix)}]\n"""

        remain = self.children_size - 1
        for i in sorted(self.children, key=lambda o: (str(o.name.prefix), o.name.name)):
            new_prefix = prefix + ("    " if not remain else "   |")
            if not prefix:
                new_prefix = new_prefix[3:]
            res += i.test_print(new_prefix, not remain)
            remain -= 1
        return res

class BuildingModel(AModel):
    __slots__ = (
        "name",
        "children",
        "yang_type",
        "units",
        "namespace",
        "default",
        "mandatory",
        "status",
        "presence_container",
        "user_ordered",
        "local_keys",
        "data_def_stm",
        "target_path",
        "identity_bases",
        "_parent",
        "config",
        "blueprint",
    )

    def __init__(self, name: Identifier, data_def_stm: AModel.StatementType, parent):
        self.name = Identifier.builtin(name) if type(name) == str else name
        self.children: List[Model] = []
        self.yang_type: Optional[YangType] = None
        self.units: Optional[str] = None
        self.namespace: str = None
        self.default: Optional[str] = None
        self.mandatory: Optional[str] = None
        self.status: Optional[str] = None
        self.presence_container = False
        self.user_ordered = False
        self.local_keys: List[str] = []
        self.data_def_stm = data_def_stm
        self.target_path = None
        self.parent = parent
        self.identity_bases = None if data_def_stm != AModel.StatementType.identity_ else []
        self.config: bool = True
        self.blueprint = []

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        self._parent = new_parent
        if new_parent:
            new_parent.children.append(self)

    @property
    def has_children(self):
        return bool(self.children)

    @property
    def children_size(self):
        return len(self.children)

    @property
    def prefix(self):
        return self.name.prefix

    @prefix.setter
    def prefix(self, v):
        self.name._prefix = v

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self is i.parent
            i.sanity_parent_child_test()

    def delete_from_parent(self, *, quiet=True):
        idx = self.parent.children.index(self)
        if idx < 0:
            if quiet:
                return
            else:
                raise make_exception(pysros_err_cannot_remove_node, node=self.name)
        del self.parent.children[idx]
        self.parent = None

    def deepcopy(self, parent):
        cls = self.__class__
        result = cls.__new__(cls)
        for i in self.__slots__:
            if i in ("_parent", "children"):
                continue
            setattr(result, i, copy.deepcopy(getattr(self, i)))
        result.parent = parent
        result.children = []
        for child in self.children:
            child.deepcopy(result)
        return result

class StorageModel(AModel):
    __slots__ = (
        "_storage",
        "_index",
    )

    class DataBitmask(IntFlag):
        data_def_stm        = 0xff
        presence_container  = 1 << 8
        user_ordered        = 1 << 9
        config              = 1 << 10
        mandatory           = 1 << 11
        status              = 1 << 12

    class DataMembers(IntEnum):
        name            = 0
        children        = 1
        yang_type       = 2
        local_keys      = 3
        target_path     = 4
        identity_bases  = 5
        parent          = 6
        bitmask         = 7
        units           = 8
        namespace       = 9
        default         = 10

class Model(StorageModel):
    """Class to represent a single YANG entry and hold additional information, such as type, configuration state, children, and data definition statement.

    .. Reviewed by TechComms 20210713
    """
    __slots__ = ("_data", "children", "parent", "name", "_bitmask", "data_def_stm",)

    def __init__(self, storage, index, parent):
        assert storage
        assert isinstance(storage, list)
        assert 0 <= index < len(storage)
        self._storage = storage
        self._index = index
        self._data = self._storage[self._index]
        self.parent = parent

    def __getattr__(self, name):
        getter = {
            "name": Model._name_getter,
            "children": Model._children_getter,
            "_bitmask": Model._bitmask_getter,
            "data_def_stm": Model._data_def_stm_getter,
        }.get(name)
        if not getter:
            raise AttributeError(name)
        res = getter(self)
        setattr(self, name, res)
        return res

    def _name_getter(self):
        name = self._data[Model.DataMembers.name]
        if name[0] is None:
            return Identifier.builtin(name[1])
        return Identifier(*name)

    def _children_getter(self):
        return [Model(self._storage, index, self) for index in self._data[Model.DataMembers.children]]

    def _bitmask_getter(self):
        return self._data[Model.DataMembers.bitmask]

    def _data_def_stm_getter(self):
        return Model.StatementType(int(self._bitmask & Model.DataBitmask.data_def_stm))

    @property
    def prefix(self):
        return self._data[Model.DataMembers.name][0]

    @property
    def has_children(self):
        return bool(self.children_data)

    @property
    def children_size(self):
        return len(self.children_data)

    @property
    def children_data(self):
        return self._data[Model.DataMembers.children]

    @property
    def yang_type(self):
        return self._data[Model.DataMembers.yang_type]

    @property
    def units(self):
        return self._data[Model.DataMembers.units]

    @property
    def namespace(self):
        return self._data[Model.DataMembers.namespace]

    @property
    def default(self):
        return self._data[Model.DataMembers.default]

    @property
    def mandatory(self):
        return bool(self._bitmask & Model.DataBitmask.mandatory)

    @property
    def status(self):
        return bool(self._bitmask & Model.DataBitmask.status)

    @property
    def presence_container(self):
        return bool(self._bitmask & Model.DataBitmask.presence_container)

    @property
    def user_ordered(self):
        return bool(self._bitmask & Model.DataBitmask.user_ordered)

    @property
    def is_region_blocked(self):
        return self.prefix in ("nokia-bof-state", "nokia-li-state", "nokia-debug-state", "nokia-li-conf", "nokia-bof-conf", "nokia-debug-conf")

    @property
    def local_keys(self):
        return self._data[Model.DataMembers.local_keys]

    @property
    def target_path(self):
        return self._data[Model.DataMembers.target_path]

    @property
    def identity_bases(self):
        return self._data[Model.DataMembers.identity_bases]

    @property
    def config(self):
        return bool(self._bitmask & Model.DataBitmask.config)

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self._data is i.parent._data
            i.sanity_parent_child_test()

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
        return f"""Model("{str(self.name)}")"""

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def __eq__(self, other):
        return self._data is other._data

class StorageConstructionModel(StorageModel):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        storage, data = new_Model_data(*args, **kwargs)
        self._storage = storage
        self._index = len(storage)
        self._storage.append(data)
        self.parent = self.parent

    @classmethod
    def new_from_data(cls, storage, index):
        assert storage
        assert isinstance(storage, list)
        assert 0 <= index < len(storage)
        res = cls.__new__(cls)
        res._storage = storage
        res._index = index
        return res

    @property
    def name(self):
        name = self._data[self.DataMembers.name]
        if name[0] is None:
            return Identifier.builtin(name[1])
        return Identifier(*name)

    @property
    def prefix(self):
        return self._data[self.DataMembers.name][0]

    @prefix.setter
    def prefix(self, value):
        self._data[self.DataMembers.name][0] = value

    @property
    def children(self):
        return (self.new_from_data(self._storage, index) for index in self._data[self.DataMembers.children])

    @property
    def has_children(self):
        return bool(self.children_data)

    @property
    def children_size(self):
        return len(self.children_data)

    @property
    def children_data(self):
        return self._data[self.DataMembers.children]

    @children_data.setter
    def children_data(self, data):
        self._data[self.DataMembers.children] = data

    @property
    def yang_type(self):
        return self._data[self.DataMembers.yang_type]

    @yang_type.setter
    def yang_type(self, value):
        self._data[self.DataMembers.yang_type] = value

    @property
    def units(self):
        return self._data[self.DataMembers.units]

    @units.setter
    def units(self, value):
        self._data[self.DataMembers.units] = value

    @property
    def namespace(self):
        return self._data[self.DataMembers.namespace]

    @namespace.setter
    def namespace(self, value):
        self._data[self.DataMembers.namespace] = value

    @property
    def default(self):
        return self._data[self.DataMembers.default]

    @default.setter
    def default(self, value):
        self._data[self.DataMembers.default] = value

    @property
    def mandatory(self):
        return bool(self._bitmask & self.DataBitmask.mandatory)

    @mandatory.setter
    def mandatory(self, value):
        if value == "true":
            self._bitmask |= int(self.DataBitmask.mandatory)
        else:
            self._bitmask &= int(~self.DataBitmask.mandatory)

    @property
    def status(self):
        return bool(self._bitmask & self.DataBitmask.status)

    @status.setter
    def status(self, value):
        if value in (None, "current", "deprecated"):
            self._bitmask |= int(self.DataBitmask.status)
        else:
            self._bitmask &= int(~self.DataBitmask.status)

    @property
    def presence_container(self):
        return bool(self._bitmask & self.DataBitmask.presence_container)

    @presence_container.setter
    def presence_container(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.presence_container)
        else:
            self._bitmask &= int(~self.DataBitmask.presence_container)

    @property
    def user_ordered(self):
        return bool(self._bitmask & self.DataBitmask.user_ordered)

    @user_ordered.setter
    def user_ordered(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.user_ordered)
        else:
            self._bitmask &= int(~self.DataBitmask.user_ordered)

    @property
    def local_keys(self):
        return self._data[self.DataMembers.local_keys]

    @local_keys.setter
    def local_keys(self, value):
        self._data[self.DataMembers.local_keys] = value

    @property
    def data_def_stm(self):
        return self.StatementType(int(self._bitmask & self.DataBitmask.data_def_stm))

    @property
    def target_path(self):
        return self._data[self.DataMembers.target_path]

    @target_path.setter
    def target_path(self, value):
        self._data[self.DataMembers.target_path] = value

    @property
    def identity_bases(self):
        return self._data[self.DataMembers.identity_bases]

    @identity_bases.setter
    def identity_bases(self, value):
        self._data[self.DataMembers.identity_bases] = value

    @property
    def parent(self):
        parent = self._data[self.DataMembers.parent]
        if parent is None:
            return parent
        return self.new_from_data(self._storage, parent)

    @parent.setter
    def parent(self, new_parent):
        self._data[self.DataMembers.parent] = new_parent if new_parent is None else new_parent._index
        if new_parent:
            assert new_parent._storage is self._storage
            new_parent.children_data.append(self._index)

    @property
    def config(self):
        return bool(self._bitmask & self.DataBitmask.config)

    @config.setter
    def config(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.config)
        else:
            self._bitmask &= int(~self.DataBitmask.config)

    @property
    def _bitmask(self):
        return self._data[self.DataMembers.bitmask]

    @_bitmask.setter
    def _bitmask(self, value):
        self._data[self.DataMembers.bitmask] = value

    @property
    def _data(self):
        return self._storage[self._index]

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self._data is i.parent._data
            i.sanity_parent_child_test()

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
        return f"""Model("{str(self.name)}")"""

    def deepcopy(self, parent):
        cls = self.__class__
        result = cls.__new__(cls)
        result._storage = parent._storage
        result._index = len(result._storage)
        result._storage.append(copy.deepcopy(self._data))
        result.children_data = []
        result.parent = parent
        for child in self.children:
            child.deepcopy(result)
        return result

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def __eq__(self, other):
        return self._data is other._data

def new_Model_data(name: Identifier, data_def_stm: Model.StatementType, parent):
    return (parent._storage if parent is not None else []), [
        [None, name] if type(name) == str else [name.prefix, name.name],        # 0 = name
        [],                                                                     # 1 = children
        None,                                                                   # 2 = yang_type
        [],                                                                     # 3 = local_keys
        None,                                                                   # 4 = target_path
        None if data_def_stm != Model.StatementType.identity_ else [],          # 5 = identity_bases
        parent if parent is None else parent._index,                            # 6 = parent
        data_def_stm.value | Model.DataBitmask.config,                          # 7 = bitmask
        None,                                                                   # 8 = units
        None,                                                                   # 9 = namespace
        None,                                                                   # 10 = default
        None,                                                                   # 11 = mandatory
        None,                                                                   # 12 = status
    ]



# Copyright 2021-2023 Nokia

from .exceptions import *
from .exceptions import make_exception

__doc__ = """This module contains error definitions for pySROS.

.. reviewed by PLM 20211201
.. reviewed by TechComms 21211202
"""

pysros_err_action_subtree_not_supported = (SrosMgmtError, """Md-compare subtree-path is not supported in action method, use compare method instead""")
pysros_err_arg_must_be_string = (TypeError, """Argument "module" must be a string""")
pysros_err_attr_cannot_be_deleted = (AttributeError, "'{obj}' object attribute '{attribute}' cannot be deleted")
pysros_err_attr_is_read_only = (AttributeError, "'{obj:.50}' object attribute '{attribute:.100}' is read-only")
pysros_err_can_check_state_from_running_only = (SrosMgmtError, "State can be checked from running datastore only")
pysros_err_can_get_state_from_running_only = (LookupError, "State can be retrieved from running datastore only")
pysros_err_can_have_one_semicolon = (InvalidPathError, "Identifier can contain only one ':'")
pysros_err_can_not_find_yang = (ModelProcessingError, "Cannot find yang '{yang_name}'")
pysros_err_cannot_call_go_to_parent = (InvalidPathError, "Cannot call go_to_parent on root")
pysros_err_cannot_delete_from_state = (SrosMgmtError, "Cannot delete from state tree")
pysros_err_cannot_lock_and_unlock_running = (SrosMgmtError, "Cannot lock and unlock running config")
pysros_err_cannot_modify_config = (SrosMgmtError, "Cannot modify running config")
pysros_err_cannot_modify_state = (SrosMgmtError, "Cannot modify state tree")
pysros_err_cannot_pars_path = (ModelProcessingError, "Cannot parse path {path!r}")
pysros_err_cannot_remove_node = (ModelProcessingError, "Cannot remove {node}")
pysros_err_cannot_specify_non_key_leaf = (InvalidPathError, "Cannot specify non-key leaf as path attribute")
pysros_err_commit_conflicts_detected = (SrosConfigConflictError, "Commit failed - conflict detected, configuration changes cleared")
pysros_err_could_not_create_conn = (RuntimeError, "Cannot create connection - {reason}")
pysros_err_data_missing = (SrosMgmtError, "Entry does not exist")
pysros_err_depth_must_be_positive = (ValueError, "Depth must be > 0")
pysros_err_duplicate_found = (SrosMgmtError, "Entry cannot contain duplicates - {duplicate}")
pysros_err_empty_path = (InvalidPathError, "Empty path")
pysros_err_entry_does_not_exists = (KeyError, "Entry does not exist")
pysros_err_even_num_of_columns_required = (ValueError, "Even number of data columns required")
pysros_err_expected_end_bracket = (InvalidPathError, "Expected ']'")
pysros_err_expected_equal_operator = (InvalidPathError, "Expected '='")
pysros_err_filter_empty_string = (SrosMgmtError, "Cannot filter by an empty string")
pysros_err_filter_not_supported_on_leaves = (InvalidPathError, "Filter is not supported for leaves")
pysros_err_filter_should_be_dict = (TypeError, "Filter argument should be a dict")
pysros_err_filter_wrong_leaf_value = (TypeError, "Unsupported leaf filter for '{leaf_name}'")
pysros_err_incorrect_leaf_value = (SrosMgmtError, "Invalid value for leaf {leaf_name}")
pysros_err_invalid_align = (ValueError, "Invalid align: '{align}'")
pysros_err_invalid_col_description = (TypeError, "Invalid column description")
pysros_err_invalid_config = (ModelProcessingError, "Invalid config statement")
pysros_err_invalid_identifier = (InvalidPathError, "Invalid identifier")
pysros_err_invalid_json_structure = (ValueError, "Invalid JSON structure")
pysros_err_invalid_key_in_path = (SrosMgmtError, "Invalid key value in path")
pysros_err_invalid_module_set_id_or_content_id = (RuntimeError, "Invalid module-set-id")
pysros_err_invalid_operation_on_key = (InvalidPathError, "Operation cannot be performed on key")
pysros_err_invalid_operation_on_leaflist = (InvalidPathError, "Operation cannot be performed on leaflist")
pysros_err_invalid_path_operation_missing_keys = (InvalidPathError, "Cannot perform operation on list without specifying keys")
pysros_err_invalid_target = (ValueError, "Invalid target")
pysros_err_invalid_transport = (TypeError, "Currently only NETCONF transport is supported")
pysros_err_invalid_value = (TypeError, "MO contents not a dict '{data}'")
pysros_err_invalid_value_for_type = (TypeError, "Invalid value for {type}: {value}")
pysros_err_invalid_yang_path = (ModelProcessingError, "Invalid path {path!r}")
pysros_err_invalid_xml_element = (ValueError, 'XML element {element} can be empty or contain another XML element')
pysros_err_invalid_xml_root = (ValueError, 'XML root element can be empty or contain another XML element')
pysros_err_key_val_mismatch = (SrosMgmtError, "Cannot change value of key-leaf '{key_name}'")
pysros_err_leaflist_should_be_list = (TypeError, "expected list object but got {type_name}")
pysros_err_malformed_keys = (TypeError, "Malformed keys for '{full_path}' with value '{value}'")
pysros_err_missing_keys = (InvalidPathError, "Missing keys on element '{element}'")
pysros_err_no_data_found = (LookupError, "No data found")
pysros_err_not_connected = (RuntimeError, "Not connected")
pysros_err_not_found_slash_before_name = (InvalidPathError, "'/' not found before element name")
pysros_err_path_should_be_string = (TypeError, "path argument should be a string")
pysros_err_prefix_does_not_have_ns = (LookupError, "prefix '{prefix}' of '{name}' does not have corresponding namespace")
pysros_err_root_path = (InvalidPathError, "Operation cannot be performed on root")
pysros_err_target_should_be_list = (InvalidPathError, "Target should be a list")
pysros_err_type_must_be = (TypeError, "must be {expected:.50s}, not {actual:.50s}")
pysros_err_unended_quoted_string = (InvalidPathError, "Unended quoted string")
pysros_err_unexpected_change_of_path_ctx = (ModelProcessingError, "Unexpected change of path ctx")
pysros_err_unexpected_end_of_yang = (ModelProcessingError, "Unexpected end of YANG")
pysros_err_unexpected_token = (ModelProcessingError, "Unexpected token {token}")
pysros_err_unexpected_value_of_type = (TypeError, "Unexpected value of type '{val_type}' in '{type}'")
pysros_err_unknown_child = (SrosMgmtError, "Cannot find child with name '{child_name}' in path '{path}'")
pysros_err_unknown_dds = (InternalError, "Unknown data definition statement type {dds}")
pysros_err_unknown_dev_statement = (ModelProcessingError, "Unknown deviate statement: {stmt!r}'")
pysros_err_unknown_element = (InvalidPathError, "Unknown element '{element}'")
pysros_err_unknown_identityref = (ModelProcessingError, "Unknown identityref {identifier}")
pysros_err_unknown_key = (InvalidPathError, "Unknown key '{key_name}'")
pysros_err_unknown_prefix_for_name = (ModelProcessingError, "Unknown prefix '{prefix}' for '{name}'")
pysros_err_unresolved_augment = (InternalError, "Augments cannot be resolved")
pysros_err_unresolved_leafref = (InternalError, "Unresolved leafref {type}")
pysros_err_unresolved_type = (InternalError, "Unresolved type in schema: {type}")
pysros_err_unsupported_set_method = (SrosMgmtError, "Unsupported set method")
pysros_err_unsupported_type_for_wrapper = (TypeError, "Unsupported type for {wrapper_name} data")
pysros_err_use_deepcopy = (NotImplementedError, "Use '.deepcopy' instead")
pysros_err_wrong_netconf_response = (SrosMgmtError, "Wrong NETCONF response")
pysros_err_wrong_rhs = (ModelProcessingError, "Invalid argument to the right of the plus symbol")
pysros_err_unsupported_compare_method = (SrosMgmtError, "Unsupported compare method")
pysros_err_unsupported_compare_datastore = (SrosMgmtError, "Compare is only supported against the candidate datastore")
pysros_err_unsupported_compare_endpoint = (InvalidPathError, "Unsupported compare path endpoint")
pysros_err_unsupported_action_path = (InvalidPathError, "Path does not point to an action")
pysros_err_unsupported_convert_method = (SrosMgmtError, "Unsupported convert method")
pysros_err_unsupported_action_io = (SrosMgmtError, "Unsupported value of argument action_io")
pysros_err_invalid_rd_state = (InternalError, "Invalid database state")
pysros_err_convert_root_not_support_pysros = (SrosMgmtError, "'pysros' format is not supported for root")
pysros_err_convert_wrong_payload_type = (TypeError, "Invalid payload type for convert")
pysros_err_wrong_json_root = (JsonDecodeError, "JSON root must be object")
pysros_err_malformed_xml = (ValueError, "Malformed XML")
pysros_err_multiple_occurences_of_node = (SrosMgmtError, "Multiple occurrences of node")
pysros_err_multiple_occurences_of_entry = (SrosMgmtError, "Multiple occurrences of list entry")
pysros_err_convert_invalid_value_for_type = (SrosMgmtError, "Invalid value for {name}")
